-- combat log parser

local CLP = {
	version = 1.0
}

function CLP:new()
	setmetatable({}, CLP)
	
	return self
end

function CLP:get_attacker(message)
	local attacker = nil

	if (self:is_attack(message) or self:is_dodged(message)) then 
		local attacks_start, attacks_end = string.find(message,  " attacks ")
		if (attacks_start != nil) then
			attacker = string.sub(message, 1, attacks_start - 1)
		else
			attacks_start, attacks_end = string.find(message,  " tried to attack ")
			if (attacks_start != nil) then
				attacker = string.sub(message, 1, attacks_start - 1)
			end
		end
	elseif (self:is_kill(message)) then 
		local split = CLP.split(message, " has been slain by ")
		local temp_attacker = split[2]:gsub("!", "")
		temp_attacker = CLP.split(temp_attacker, "'s")
		attacker = temp_attacker[1]
	end

	return attacker
end

function CLP:get_attacked(message) 
	local attacked = nil

	if (self:is_attack(message) or self:is_dodged(message)) then 
		local attacks_start, attacks_end = string.find(message, " attacks ")
		local hits_start, hits_end = string.find(message, " and ")
		if (attacks_start != nil and hits_start != nil) then
			attacked = string.sub(message, attacks_end + 1, hits_start - 1)
		else
			local attacks_start, attacks_end = string.find(message, " tried to attack ")
			local hits_start, hits_end = string.find(message, " but ")
			if (attacks_start != nil and hits_start != nil) then
				attacked = string.sub(message, attacks_end + 1, hits_start - 1)
			end
		end
	elseif (self:is_kill(message)) then 
		local split = CLP.split(message, " has been slain by ")
		attacked = split[1]
	end

	return attacked
end

function CLP:get_skill(message)
	local skill = nil

	if (self:is_attack(message)) then 
		local check_start, check_end = string.find(message,  " from ")
		if (check_start != nil) then
			local period_start, period_end = string.find(message,  "%.", check_end + 1)
			if (period_start != nil) then 
				skill = string.sub(message, check_end + 1, period_start - 1)
			else
				skill = string.sub(message, check_end + 1)
			end
		end
	elseif (self:is_kill(message)) then 
		local check_start, check_end = string.find(message,  " by ")
		if (check_start != nil) then
			local skill_split = CLP.split(message, "'s ")
			if (skill_split[2] ~= nil) then 
				skill = skill_split[2]:gsub("!", "")
			end
		end
	end
	return skill
end

function CLP:get_damage_amount(message)
	local damage = 0

	local check_start, check_end = string.find(message,  " dealing ")
	if (check_start != nil) then
		local points_start, points_end = string.find(message,  " points ", check_end + 1)
		if (points_start != nil) then 
			damage = tonumber(string.sub(message, check_end + 1, points_start - 1))
		end
	end

	return damage
end

function CLP:is_player(check_name)
	local is_player = false
	
	if (check_name != nil) then 
		is_player = (string.upper(check_name) == string.upper(player.name))
	end
	
	return is_player
end

function CLP:is_party_member(check_name)
	local is_party_member = false
	
	if (check_name != nil) then				
		is_party_member = (ShroudGetPartyMemberMaxFocusInScene(check_name) > -1)
	end
	
	return is_party_member
end

function CLP:is_pet(check_name)
	return self:get_pet_name(check_name) != nil
end

function CLP:get_pet_owner(check_name)
	local pet_owner = nil
	
	if (self:is_pet(check_name)) then
		local owner_start_start, owner_start_end = string.find(check_name, "%<")
		if (owner_start_end != nil) then
			local owner_end_start, owner_end_end = string.find(check_name, "%>")
			if (owner_end_start != nil) then
				pet_owner = string.sub(check_name, owner_start_end + 1, owner_end_start -1)
			end
		end
	end
	
	return pet_owner
end

function CLP:get_pet_name(check_name)
	local pet_name = nil
	
	if (check_name != nil) then 
		local owner_start_start, owner_start_end = string.find(check_name, "%<")
		if (owner_start_start != nil) then
			pet_name = string.sub(check_name, 1, owner_start_start - 2)
		end
	end 
	
	return pet_name
end

function CLP:is_kill(message)
	local is_kill = false
	
	local kill_start, kill_end = string.find(message, " has been slain by ")
	if kill_start != nil then 
		is_kill = true
	end
	
	return is_kill
end

function CLP:is_attack(message)
	local is_attack = false
	
	local attacks_start, attacks_end = string.find(message, " attacks ")
	if attacks_start != nil then 
		is_attack = true
	end
	
	return is_attack
end

function CLP:is_hit(message)
	local is_hit = false
	
	local check_start, check_end = string.find(message, " and hits")
	if check_start != nil then 
		is_hit = true
	end
	
	return is_hit
end

function CLP:is_dodged(message)
	local is_dodged = false
	
	local check_start, check_end = string.find(message, " dodges")
	if check_start != nil then 
		is_dodged = true
	end
	
	return is_dodged
end

function CLP:is_critical(message)
	local is_critical = false
	
	local check_start, check_end = string.find(message, " of critical damage")
	if check_start != nil then 
		is_critical = true
	end
	
	return is_critical
end

function CLP:is_glancing(message)
	local is_glancing = false
	
	local check_start, check_end = string.find(message, " of glancing damage ")
	if check_start != nil then 
		is_glancing = true
	end
	
	return is_glancing			
end

function CLP:is_blocked(message)
	local is_blocked = false
	
	local check_start, check_end = string.find(message, " and is blocked")
	if check_start != nil then 
		is_blocked = true
	end
	
	return is_blocked			
end

function CLP:is_parried(message)
	local is_parried = false
	
	local check_start, check_end = string.find(message, " past the parry")
	if check_start != nil then 
		is_parried = true
	end
	
	return is_parried			
end

function CLP:is_healed(message)
	local is_healed = false
	
	local check_start, check_end = string.find(message, " is healed ")
	if check_start != nil then 
		is_healed = true
	end
	
	return is_healed			
end

function CLP:get_healed(message)
	local healed = nil
			
	if (self:is_healed(message)) then 
		local check_start, check_end = string.find(message,  " is healed ")
		if (check_start != nil) then
			healed = string.sub(message, 1, check_start - 1)
		end
	end

	return healed			
end

function CLP:get_healer(message)
	local healer = nil

	if (self:is_healed(message)) then 
		local check_start, check_end = string.find(message,  " of health by ")
		if (check_start != nil) then
			local period_start, period_end = string.find(message,  '%.', check_end + 1)
			if (period_start != nil) then 
				healer = string.sub(message, check_end + 1, period_start - 1)
			else
				healer = string.sub(message, check_end + 1)
			end
		end
	end

	return healer			
end

function CLP:get_heal_amount(message)
	local heal = 0

	local check_start, check_end = string.find(message,  " for ")
	if (check_start != nil) then
		local points_start, points_end = string.find(message,  " points ", check_end + 1)
		if (points_start != nil) then 
			heal = tonumber(string.sub(message, check_end + 1, points_start - 1))
		end
	end
	
	return heal
end

-- https://stackoverflow.com/questions/1426954/split-string-in-lua
function CLP.split(s, delimiter)
    result = {}
    for match in (s..delimiter):gmatch("(.-)"..delimiter) do
        table.insert(result, match)
    end
    return result
end

return CLP