local PQS = {
	version = ".6a",
	settings_file = "user.ini",
	player_settings = {},
	LIP = nil,
	CLP = nil,
	split = nil,
	window_system = nil,	
	PQS_WINDOW = nil,
	event_timer = nil,
	quests = {},
	current_distance_check = {},
	current_pet_name = "",
	current_matched_criteria_messages = {},
	last_distance_check = {},
	last_scene_name = "",
	last_pet_name = "",
	last_timestamp = nil,
	console_log_message_queue = {},
	menu_button_width = 165,
	menu_button_spacing = 30,
	tiny_tex = nil, 
	access_button = nil,
	scale = 1.25,
	alert_icon = nil
}

function PQS.load_player_settings()	
	
	-- ensure the user setting file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath..'/pqs/' .. PQS.settings_file, "r")
	if file == nil then 
		file = io.open(ShroudLuaPath..'/pqs/' .. PQS.settings_file, "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end
	
	PQS.player_settings = PQS.LIP().load(ShroudLuaPath..'/pqs/' .. PQS.settings_file)

	if (PQS.player_settings[player.name .. "|access_button"] == nil) then
		PQS.player_settings[player.name .. "|access_button"] = {}
		PQS.player_settings[player.name .. "|access_button"]['scale'] = PQS.scale
		PQS.player_settings[player.name .. "|access_button"]['left'] = client.screen.width - (50)
		PQS.player_settings[player.name .. "|access_button"]['top'] = client.screen.height - (50)
		PQS.save_player_settings()
	end	
end

function PQS.save_player_settings()
	PQS.LIP().save(ShroudLuaPath..'/pqs/' .. PQS.settings_file, PQS.player_settings)
end

function PQS.scale_text(text_to_scale, font_size, scale)
	return "<size=" .. (font_size * scale) .. ">" .. text_to_scale .. "</size>"
end

function PQS.initialize()
	math.randomseed(os.time())

	PQS.LIP = ui.module.add("pqs/modules/LIP.lua")
	PQS.CLP = ui.module.add("pqs/modules/CLP.lua")
	PQS.CLP = PQS.CLP():new()
	PQS.split = ui.module.add("pqs/modules/split.lua")
	PQS.split = PQS.split()
	
	PQS.load_player_settings()

	PQS.scale = PQS.player_settings[player.name .. "|access_button"]['scale']

	PQS.window_system = ui.module.add("pqs/modules/PQS_WINDOW_SYSTEM.lua")
	PQS.window_system = PQS.window_system()
	PQS.window_system:initialize()

	PQS.PQS_WINDOW = ui.module.add("pqs/modules/PQS_WINDOW.lua")
	PQS.tiny_tex = ui.texture.add(0, 0, "pqs/modules/tiny.png")
	PQS.alert_icon = ui.texture.add(0, 0, "pqs/modules/alert_icon.png")

	-- create the timers for the system
	PQS.create_timers()

	PQS.last_timestamp = PQS.get_server_timestamp()

	-- time in UTC
	--ConsoleLog(PQS.last_timestamp)
	
	--local datetime = os.date('!*t')
	--print(datetime.year, datetime.month, datetime.day)
	
	PQS.build_windows()

	PQS.build_access_button()
	
	PQS.move_access_button()
	
	ui.onUpdate(function()	
		PQS.access_button:visible(ShroudIsUIActive())				
	end)

	ui.onConsoleInput(function(channel, sender, dst, msg)
		--PQS_system.process_log_message(channel, sender, dst, msg)
		--print(channel, sender, dst, msg)
	end)
	
	ConsoleLog("[808080]-------------------------[-]")
	ConsoleLog("[808080]-------------------------[-]")
	ConsoleLog("Welcome to version " .. PQS.version .. " of the Player Quest System lua addon for SotA")
	ConsoleLog("This lua addon can be found [FFFF00][url=https://www.shroudoftheavatar.com/forum/index.php?threads/player-quest-system-wip.169997/]here[/url][-] on the SotA forum.")
	ConsoleLog("This lua addon requires [FFFF00][url=https://www.shroudoftheavatar.com/forum/index.php?threads/libsota-a-helper-library-in-lua-for-shroud-in-lua.160222/]libsota[/url][-] version 0.4.6+.  Version 0.4.8 is included with BMS.")
	ConsoleLog("[808080]-------------------------[-]")
	ConsoleLog("Hold [FFFF00]LeftControl[-] and [FFFF00]click[-] the PQS icon to move.  Release the LeftControl key when you are done moving.")
	ConsoleLog("[808080]-------------------------[-]")
	
end

function PQS.create_timers()
	PQS.event_timer = ui.timer.add(1, false, function()
	
		local pet_info = ShroudGetPetInfo()

		if (pet_info ~= nil) then
			PQS.current_pet_name = pet_info.name
		else
			PQS.current_pet_name = ""
		end
		
		PQS.last_distance_check = PQS.current_distance_check
		PQS.current_distance_check = {}
		
		for quest_id, quest in pairs(PQS.quests) do			
			if (PQS.is_quest_enabled(quest_id)) then 
				
				PQS.update_player_timers(quest)
			
				PQS.check_locations(quest_id, quest)
				
				PQS.check_start_events(quest_id, quest)

				PQS.check_events(quest_id, quest)
				
				quest:set_last_player_data()
			end
		end

		PQS.refresh_alert()
		
		PQS.last_scene_name = player.location.scene.name
						
		PQS.last_pet_name = PQS.current_pet_name
		
		PQS.console_log_message_queue = {}
		
		PQS.current_matched_criteria_messages = {}
		
		PQS.last_timestamp = PQS.get_server_timestamp()
		
	end)
end

function PQS.update_player_timers(quest)
	local active_timer_list = quest:get_active_timer_list(player.name)

	for timer_index, timer_id in pairs(active_timer_list) do
		quest:update_timer_ticks(player.name, timer_id)
	end
	
end

function PQS.check_start_events(quest_id, quest)
	if (PQS.get_quest_variable(quest_id, 'quest_started') ~= 1) then
	
		PQS.set_quest_variable(quest_id, 'quest_started', 1)
		PQS.on_quest_started(quest)

		-- if there is a start journal id, add it
		local start_journal_id = PQS.get_quest_data(quest, 'start_journal')
		if (start_journal_id ~= nil) then
			PQS.add_journal_entry(quest, start_journal_id)
		end
	end
end

function PQS.add_journal_entry(quest, journal_id)
	quest:add_journal_entry(player.name, journal_id)
	PQS.quest_consolelog(quest, "A journal entry has been added")
	PQS.show_alert()
end

function PQS.on_quest_started(quest)
	PQS.quest_consolelog(quest, "Started")
end

function PQS.check_locations(quest_id, quest)
		
	local location_list = quest:get_location_list()
	
	for location_key, location_id in pairs(location_list) do 					
		local location = quest:get_location(location_id)
		if (location ~= nil and location['scene'] == player.location.scene.name) then				
			local distance = PQS.distance(location['x'], location['y'], location['z'], player.location.x, player.location.y, player.location.z)
			PQS.current_distance_check[location_id] = distance
		end
	end	
end

-- ---------------------------------------------------------
-- Event functions
-- ---------------------------------------------------------

-- checks the event criteria to see if it can be processed
function PQS.can_process_event_actions(quest, event)

	-- TODO: criteria for has/has not viewed dialog

	local quest_id = quest:get_id()	
	local can_process_event_actions = false;
	
	PQS.current_matched_criteria_messages = {}

	if (quest ~= nil and event ~= nil) then
	
		local all_criteria_good = true

		-- make sure this event has criteria
		local criteria_list = quest:get_event_criteria_list(event['id'])
					
		-- check each criteria assigned to the event
		for criteria_key, criteria_id in pairs(criteria_list) do
		
			local criteria = quest:get_event_criteria(event['id'], criteria_id)
			
			if (criteria ~= nil) then
				if (criteria['type'] == 'timestamp') then 
				
					local current_timestamp = PQS.get_server_timestamp()
				
					if (criteria['comparison'] == 'less') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] < current_timestamp 
						)
					elseif (criteria['comparison'] == 'greater') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] > current_timestamp 
						)
					end
					
				elseif (criteria['type'] == 'scene') then 
					if (criteria['comparison'] == 'is') then 
						all_criteria_good = (all_criteria_good and criteria['value'] == player.location.scene.name)
						
					elseif (criteria['comparison'] == 'is_not') then 
						all_criteria_good = (all_criteria_good and criteria['value'] ~= player.location.scene.name)

					elseif (criteria['comparison'] == 'enter') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] == player.location.scene.name and 
							PQS.last_scene_name ~= criteria['value']
						)
						
					elseif (criteria['comparison'] == 'exit') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] ~= player.location.scene.name and 
							PQS.last_scene_name == criteria['value']
						)
						
					end
									
				elseif (criteria['type'] == 'pet') then 
					if (criteria['value'] == nil) then 
						criteria['value'] = ""
					end
				
					if (criteria['comparison'] == 'is') then 
						all_criteria_good = (all_criteria_good and ((criteria['value'] == "" and PQS.current_pet_name ~= "") or string.match(PQS.current_pet_name, criteria['value']) == criteria['value']))
						
					elseif (criteria['comparison'] == 'is_not') then 
						all_criteria_good = (all_criteria_good and ((criteria['value'] == "" and PQS.current_pet_name == "") or string.match(PQS.current_pet_name, criteria['value']) ~= criteria['value']))

					elseif (criteria['comparison'] == 'enter') then 
					
						if (criteria['value'] == "") then
							all_criteria_good = (
								all_criteria_good and 
								(PQS.last_pet_name == "" and PQS.current_pet_name ~= "")
							)
						else 						
							all_criteria_good = (
								all_criteria_good and 
								string.match(PQS.current_pet_name, criteria['value']) == criteria['value'] and 
								string.match(PQS.last_pet_name, criteria['value']) ~= criteria['value']
							)
						end
						
					elseif (criteria['comparison'] == 'exit') then 
						if (criteria['value'] == "") then
							all_criteria_good = (
								all_criteria_good and 
								(PQS.last_pet_name ~= "" and PQS.current_pet_name == "")
							)
						else 						
							all_criteria_good = (
								all_criteria_good and 
								string.match(PQS.current_pet_name, criteria['value']) ~= criteria['value'] and 
								string.match(PQS.last_pet_name, criteria['value']) == criteria['value']
							)
						end
					end
				elseif (criteria['type'] == 'journal_entry') then

					-- has, has_not, viewed, not_viewed, enter
				
					if (criteria['comparison'] == 'has') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:has_journal_entry(player.name, criteria['value']) == true
						)
					elseif (criteria['comparison'] == 'has_not') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:has_journal_entry(player.name, criteria['value']) == false
						)
					elseif (criteria['comparison'] == 'has_viewed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_viewed(player.name, criteria['value']) == true
						)
					elseif (criteria['comparison'] == 'has_not_viewed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_viewed(player.name, criteria['value']) == false
						)
					elseif (criteria['comparison'] == 'viewed') then
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_viewed(player.name, criteria['value']) == true and 
							quest:is_journal_entry_viewed(player.name, criteria['value'], false, true) == false
						)
					elseif (criteria['comparison'] == 'has_completed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_complete(player.name, criteria['value']) == true
						)
					elseif (criteria['comparison'] == 'has_not_completed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_complete(player.name, criteria['value']) == false
						)
					elseif (criteria['comparison'] == 'completed') then
						all_criteria_good = (
							all_criteria_good and 
							quest:is_journal_entry_complete(player.name, criteria['value']) == true and 
							quest:is_journal_entry_complete(player.name, criteria['value'], false, true) == false
						)
					elseif (criteria['comparison'] == 'has_failed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:set_journal_entry_fail(player.name, criteria['value']) == true
						)
					elseif (criteria['comparison'] == 'has_not_failed') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:set_journal_entry_fail(player.name, criteria['value']) == false
						)
					elseif (criteria['comparison'] == 'failed') then
						all_criteria_good = (
							all_criteria_good and 
							quest:set_journal_entry_fail(player.name, criteria['value']) == true and 
							quest:set_journal_entry_fail(player.name, criteria['value'], false, true) == false
						)
					end
					
				elseif (criteria['type'] == 'dialog') then 		

					if (criteria['comparison'] == 'has') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_dialog_viewed(player.name, criteria['value'])
						)
					elseif (criteria['comparison'] == 'has_not') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:is_dialog_viewed(player.name, criteria['value']) == false
						)
					elseif (criteria['comparison'] == 'enter') then
					
						all_criteria_good = (
							all_criteria_good and 
							quest:is_dialog_viewed(player.name, criteria['value']) and 
							quest:is_dialog_viewed(player.name, criteria['value'], false, true) == false
						)
					end
				elseif (criteria['type'] == 'quest_variable') then 		

					if (criteria['comparison'] == 'is') then 
						all_criteria_good = (all_criteria_good and criteria['value'] == PQS.get_quest_variable(quest_id, criteria['name']))
					elseif (criteria['comparison'] == 'is_not') then 
						all_criteria_good = (all_criteria_good and criteria['value'] ~= PQS.get_quest_variable(quest_id, criteria['name']))
					elseif (criteria['comparison'] == 'exit') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] ~= PQS.get_quest_variable(quest_id, criteria['name']) and 
							criteria['value'] == PQS.get_quest_variable(quest_id, criteria['name'], false, true)
						)
					elseif (criteria['comparison'] == 'enter') then 						
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] == PQS.get_quest_variable(quest_id, criteria['name']) and 
							criteria['value'] ~= PQS.get_quest_variable(quest_id, criteria['name'], false, true)
						)
					elseif (criteria['comparison'] == 'less') then 
						all_criteria_good = (
							all_criteria_good and 
							PQS.get_quest_variable(quest_id, criteria['name']) < criteria['value']
						)
					elseif (criteria['comparison'] == 'greater') then 
						all_criteria_good = (
							all_criteria_good and 
							PQS.get_quest_variable(quest_id, criteria['name']) > criteria['value']
						)
					end

				elseif (criteria['type'] == 'quest_item') then 		

					if (criteria['comparison'] == 'is') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] == quest:get_player_quest_item_count(player.name, criteria['name'])
						)
					elseif (criteria['comparison'] == 'is_not') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] ~= quest:get_player_quest_item_count(player.name, criteria['name'])
						)
					elseif (criteria['comparison'] == 'exit') then 
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] ~= quest:get_player_quest_item_count(player.name, criteria['name']) and 
							criteria['value'] == quest:get_player_quest_item_count(player.name, criteria['name'], false, true)
						)
					elseif (criteria['comparison'] == 'enter') then 						
						all_criteria_good = (
							all_criteria_good and 
							criteria['value'] == quest:get_player_quest_item_count(player.name, criteria['name']) and 
							criteria['value'] ~= quest:get_player_quest_item_count(player.name, criteria['name'], false, true)
						)
					elseif (criteria['comparison'] == 'less') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:get_player_quest_item_count(player.name, criteria['name']) < criteria['value']
						)
					elseif (criteria['comparison'] == 'greater') then 
						all_criteria_good = (
							all_criteria_good and 
							quest:get_player_quest_item_count(player.name, criteria['name']) > criteria['value']
						)
					end
					
				elseif (criteria['type'] == 'chance') then 
					local chance = criteria['value']
					if (chance == nil) then chance = 100 end					
					all_criteria_good = (all_criteria_good and (chance >= math.random(1, 100)))

				elseif (criteria['type'] == 'location') then 
					if (all_criteria_good) then 
						local location = quest:get_location(criteria['value'])
						local distance_good = false
						
						if (location ~= nil and location['scene'] == player.location.scene.name) then
							if (criteria['comparison'] == 'is') then 
								distance_good = (PQS.current_distance_check[criteria['value']] <= criteria['distance'])
							elseif (criteria['comparison'] == 'is_not') then 
								distance_good = (PQS.current_distance_check[criteria['value']] > criteria['distance'])
							elseif (criteria['comparison'] == 'enter') then
								distance_good = (
									PQS.current_distance_check[criteria['value']] <= criteria['distance'] and 
									(PQS.last_distance_check[criteria['value']] == nil or PQS.last_distance_check[criteria['value']] > criteria['distance'])
								)								
							elseif (criteria['comparison'] == 'exit') then 
								distance_good = (
									PQS.current_distance_check[criteria['value']] > criteria['distance'] and 
									(PQS.last_distance_check[criteria['value']] ~= nil and PQS.last_distance_check[criteria['value']] <= criteria['distance'])
								)
							end
						end
					
						all_criteria_good = (
							all_criteria_good and 
							distance_good
						)

					end						
				elseif (criteria['type'] == 'npc') then 
				
					local found_match = false

					for message_queue_key, message in pairs(PQS.console_log_message_queue) do 
					
						if (message.message_type == 'NPC') then 
							local name = nil
						
							if (criteria['comparison'] == 'from') then
								name = message.source
							elseif (criteria['comparison'] == 'to') then 
								name = message.destination
							end
							
							-- substitute the player name
							local check_name = string.gsub(criteria['name'], '{player}', player.name)
													
							if (string.match(name, check_name) == check_name) then
								if (criteria['value'] == nil or string.match(message.message, criteria['value']) == criteria['value']) then
									-- store the message for use by the actions if needed
									table.insert(PQS.current_matched_criteria_messages, message)
									found_match = true
									break
								end
							end
						end
					end						
					
					all_criteria_good = (
						all_criteria_good and 
						found_match
					)						
				elseif (criteria['type'] == 'craft') then 
				
					local found_match = false
				
					for message_queue_key, message in pairs(PQS.console_log_message_queue) do 						
						if (message.message_type == 'Craft') then 
							local name = nil

							if (criteria["recipe"] == message.recipe) then 
								-- store the message for use by the actions if needed
								table.insert(PQS.current_matched_criteria_messages, message)
								found_match = true
								break
							end
						end
					end						
					
					all_criteria_good = (
						all_criteria_good and 
						found_match
					)						
				elseif (criteria['type'] == 'loot') then 
				
					local found_match = false
				
					for message_queue_key, message in pairs(PQS.console_log_message_queue) do 						
						if (message.message_type == 'Loot') then 																
							if (string.match(message.item, criteria['value']) == criteria['value']) then
								--print(message.quantity, criteria['quantity'], criteria['comparison'])
								if (criteria['comparison'] == nil or 
									(criteria['comparison'] == 'is' and message.quantity == criteria['quantity']) or 
									(criteria['comparison'] == 'is_not' and message.quantity ~= criteria['quantity']) or 
									(criteria['comparison'] == 'less' and message.quantity < criteria['quantity']) or 
									(criteria['comparison'] == 'greater' and message.quantity > criteria['quantity']) 
								) then								
									-- store the message for use by the actions if needed
									table.insert(PQS.current_matched_criteria_messages, message)
									found_match = true
									break
								end
							end
						end
					end						
					
					all_criteria_good = (
						all_criteria_good and 
						found_match
					)						
				elseif (criteria['type'] == 'combat') then 
					local found_match = false
					local can_continue = true
				
					for message_queue_key, message in pairs(PQS.console_log_message_queue) do 						
						if (message.message_type == 'Combat') then 
							--print(message.message)

							local attacker = PQS.CLP:get_attacker(message.message)
							local attacked = PQS.CLP:get_attacked(message.message)
							local skill = PQS.CLP:get_skill(message.message)
							
							-- check who (player, pet, party, partypet, exact name of mob, player, or pet)
							can_continue = can_continue and (
								(criteria['who'] == nil) or  
								(
									(criteria['who'] == "player" and PQS.CLP:is_player(attacker)) or  
									(criteria['who'] == "pet" and PQS.CLP:is_pet(attacker) and PQS.CLP:is_player(PQS.CLP:get_pet_owner(attacker))) or 
									(criteria['who'] == "party" and PQS.CLP:is_party_member(attacker)) or 
									(criteria['who'] == "partypet" and PQS.CLP:is_pet(attacker) and PQS.CLP:is_party_member(PQS.CLP:get_pet_owner(attacker))) or 
									(string.find(message.message,  criteria['who']))
								)
							)

							can_continue = can_continue and (
								(criteria['what'] == nil) or  
								(
									(criteria['what'] == "player" and PQS.CLP:is_player(attacker)) or  
									(criteria['what'] == "pet" and PQS.CLP:is_pet(attacker) and PQS.CLP:is_player(PQS.CLP:get_pet_owner(attacker))) or 
									(criteria['what'] == "party" and PQS.CLP:is_party_member(attacker)) or 
									(criteria['what'] == "partypet" and PQS.CLP:is_pet(attacker) and PQS.CLP:is_party_member(PQS.CLP:get_pet_owner(attacker))) or 
									(string.find(message.message,  criteria['what']))
								)
							)

							can_continue = can_continue and (
								(criteria['skill'] == nil) or  
								(
									(criteria['skill'] == skill) or 
									(criteria['skill'] == "Base Attack" and skill == nil)
								)
							)
							
							if (can_continue and criteria['comparison'] == "kill" and PQS.CLP:is_kill(message.message)) then 
																
								if (can_continue) then
									table.insert(PQS.current_matched_criteria_messages, message)
									found_match = true									
								end
							elseif (can_continue and criteria['comparison'] == "damage" and PQS.CLP:is_hit(message.message)) then
								local damage_amount = PQS.CLP:get_damage_amount(message.message)
								local is_critical = PQS.CLP:is_critical(message.message)

								can_continue = can_continue and (
									(criteria['damage'] == nil) or  
									(
										(damage_amount >= criteria['damage'])
									)
								)
								can_continue = can_continue and (
									(criteria['critical'] == nil) or is_critical  
								)
								
								if (can_continue) then
									table.insert(PQS.current_matched_criteria_messages, message)
									found_match = true									
								end
							end
						end
					end
					
					all_criteria_good = (
						all_criteria_good and 
						found_match
					)						
				end
			else
				all_criteria_good = false
			end				
		end
		
		can_process_event_actions = all_criteria_good			
	end
	
	return can_process_event_actions
end

-- processes the event actions
function PQS.process_event_actions(quest, event)

	local quest_id = quest:get_id()
	local action_list = quest:get_event_action_list(event['id'])

	-- perform each action assigned to the event
	for action_key, action_id in pairs(action_list) do
			
		local event_action = quest:get_event_action(event['id'], action_id)
		
		if (event_action ~= nil) then
		
			if event_action['type'] == 'set_quest_variable' then
			
				local set_type = 'set'
				
				if (event_action['set_type'] ~= nil) then 
					set_type = event_action['set_type']
				end
			
				if (set_type == 'set') then 
					PQS.set_quest_variable(quest_id, event_action['name'], event_action['value'])
				elseif (set_type == 'increment') then
					local current_value = PQS.get_quest_variable(quest_id, event_action['name'])
					if (current_value == nil) then 
						current_value = 0
					end

					current_value = current_value + event_action['value']
					
					PQS.set_quest_variable(quest_id, event_action['name'], current_value)
				elseif (set_type == 'decrement') then 
					local current_value = PQS.get_quest_variable(quest_id, event_action['name'])
					if (current_value == nil) then 
						current_value = 0
					end

					current_value = current_value - event_action['value']
					
					PQS.set_quest_variable(quest_id, event_action['name'], current_value)
				end
			elseif event_action['type'] == 'add_journal_entry' then
				if (quest:has_journal_entry(player.name, event_action['value']) == false) then 
					quest:add_journal_entry(player.name, event_action['value'])
					PQS.quest_consolelog(quest, "A journal entry has been added.")
				end
			elseif event_action['type'] == 'set_journal_entry_status' then
				if (quest:has_journal_entry(player.name, event_action['value']) == true) then 
					if (event_action['set_type'] == 'fail') then 
						quest:set_journal_entry_fail(player.name, event_action['value'])
						PQS.quest_consolelog(quest, "A journal task has failed.", "FF0000")
					elseif (event_action['set_type'] == 'complete') then
						quest:set_journal_entry_complete(player.name, event_action['value'])
						PQS.quest_consolelog(quest, "A journal task has been completed.", "00FF00")
					end
				end
			elseif event_action['type'] == 'console_text' then
				PQS.quest_consolelog(quest, event_action['value'])
			elseif event_action['type'] == 'start_dialog' then
				PQS.build_dialog_window(quest_id, event_action['value'])
			elseif event_action['type'] == 'timer' then
				if (event_action['action'] == 'start') then 
					if (quest:is_timer(player.name, event_action['timer_id']) == false) then 
						quest:create_timer(player.name, event_action['timer_id'])
					end
					quest:start_timer(player.name, event_action['timer_id'])
				elseif (event_action['action'] == 'stop') then 
					quest:stop_timer(player.name, event_action['timer_id'])
				elseif (event_action['action'] == 'reset') then 
					quest:reset_timer(player.name, event_action['timer_id'])
				end
			elseif event_action['type'] == 'alert_icon' then
				if (event_action['value'] == "show") then
					PQS.show_alert()
				elseif (event_action['value'] == "hide") then
					PQS.hide_alert()
				end
			elseif event_action['type'] == 'quest_item' then
				local chance = event_action['chance']
				if (chance == nil) then chance = 100 end					
				local chance_success = (chance >= math.random(1, 100))
			
				if (chance_success) then 					
					if (event_action['set_type'] == 'set') then
						quest:set_player_quest_item(player.name, event_action['name'], event_action['value'])
					elseif (event_action['set_type'] == 'increment') then
						quest:add_player_quest_item(player.name, event_action['name'], event_action['value'])
					elseif (event_action['set_type'] == 'decrement') then 
						quest:remove_player_quest_item(player.name, event_action['name'], event_action['value'])
					end
				end
			end
		end
	end
	
	--print('processing event actions')
end

function PQS.get_server_timestamp()

	local time_pattern = "(%d+)/(%d+)/(%d+) (%d+):(%d+):(%d+) (%S+)"
	local month, day, year, hour, minute, seconds, ampm = string.match(ShroudServerTime, time_pattern)

	if (ampm == "PM") then 
		hour = hour + 12
	end

	if hour == 24 then hour = 00 end

	--print(month, day, year, hour, minute, seconds, ampm)

	-- UTC timestamp
	local server_timestamp = os.time({year = year, month = month, day = day, hour = hour, min = minute, sec = seconds})
	
--[[
	local new_time = os.date("!*t", server_timestamp)
	
	for k,v in pairs(new_time) do
		print(k, v)
	end
]]
	--print(server_timestamp)

	return server_timestamp
end

-- check events 
function PQS.check_events(quest_id, quest)

	if (PQS.is_quest_enabled(quest_id)) then

		local event_list = quest:get_event_list()			
		for event_key, event_id in pairs(event_list) do 
		
			local event = quest:get_event(event_id)
			
			if (PQS.can_process_event_actions(quest, event)) then
				PQS.process_event_actions(quest, event)
			end
		end
	end
end

-- ---------------------------------------------------------
-- Quest functions
-- ---------------------------------------------------------

function PQS.get_quest_count()
	local count = 0
	for quest_id, quest in pairs(PQS.quests) do
		count = count + 1
	end
	return count
end

function PQS.get_quest_by_index(quest_index)
	local current_index = 1
	local found_quest = nil
	
	for quest_id, quest in pairs(PQS.quests) do
		if current_index == quest_index then 
			found_quest = quest
			break
		end
		
		current_index = current_index + 1
	end
	
	return found_quest
end

function PQS.get_quest(quest_id)
	local quest = nil
	if (PQS.quests[quest_id] ~= nil) then
		quest = PQS.quests[quest_id]
	end
	return quest
end

function PQS.get_quest_data(quest, data_name)
	local data_value = nil

	if (quest.quest_data['quest'][data_name] ~= nil) then 
		data_value = quest.quest_data['quest'][data_name]
	end

	return data_value
end

function PQS.set_quest_variable(quest_id, variable_name, variable_value)
	PQS.quests[quest_id]:set_quest_variable(player.name, variable_name, variable_value)
end

function PQS.get_quest_variable(quest_id, variable_name, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false

	local quest_variable_value = PQS.quests[quest_id]:get_quest_variable(player.name, variable_name, use_previous, use_last)
	return quest_variable_value
end

-- adds a quest to the available quest list
function PQS.add_quest(quest)
	if (quest ~= nil) then 
		PQS.quests[quest.quest_data['quest']['id']] = quest
		PQS.quest_consolelog(quest, "Loaded", "00FF00")
		--quest:dump()
	end
end

function PQS.quest_consolelog(quest, consolelog_message, message_color)
	message_color = message_color or "FFFFFF"
	if (quest ~= nil) then 
		ConsoleLog("[FFA500]Quest[-] [69BBF7]" .. quest.quest_data['quest']['name'] .. "[-]: [" .. message_color .. "]" .. consolelog_message .. "[-]")
	end
end

function PQS.disable_quest(quest_id)
	PQS.set_quest_variable(quest_id, 'enabled', 0)
	PQS.quest_consolelog(PQS.quests[quest_id], "Disabled", "FF0000")
end

function PQS.enable_quest(quest_id)
	PQS.set_quest_variable(quest_id, 'enabled', 1)
	PQS.quest_consolelog(PQS.quests[quest_id], "Enabled", "00FF00")
end

function PQS.is_quest_enabled(quest_id)
	local enabled = PQS.get_quest_variable(quest_id, 'enabled')
	if (enabled == nil) then
		PQS.enable_quest(quest_id)
		enabled = 1
	end
	return (enabled == 1)
end

function PQS.dump_quests()
	for quest_id, quest in pairs(PQS.quests) do 
		quest:dump()
	end
end

-- ---------------------------------------------------------
-- Window UI functions
-- ---------------------------------------------------------

function PQS.build_windows()
	PQS.build_menu()
	PQS.build_quests_window()
	PQS.build_settings_window()
end

function PQS.build_confirm_window(title, message, button_yes_text, button_no_text, data, response_function)
	if PQS.window_system:window_exists("confirm") == true then 
		PQS.window_system:destroy_window("confirm")
	end

	local window = PQS.PQS_WINDOW():new(title, 500, 100, PQS.scale, false, false, true)

	window.rebuild_content = function()
		local message_label = ui.label.add(0, 0, (window.original_width - 25), 50, string.gsub(message, "<br>", "\n"))
		window:add_element("message-label", message_label, 15, 45, false, false, 18)

		local yes_button = ui.guiButton.add(0, 0, PQS.tiny_tex, button_yes_text, "", 100, 30)
		window:add_element("yes-button", yes_button, ((window.original_width) / 2) - (100), (window.original_height), false, false, 18)

		yes_button:onClick(function(x, y) 
			PQS.handle_button_click(window.window_system_key, function()
				response_function(data, button_yes_text)
				window:close()
			end)
		end)	
		
		local no_button = ui.guiButton.add(0, 0, PQS.tiny_tex, button_no_text, "", 100, 30)
		window:add_element("no-button", no_button, ((window.original_width) / 2) + (20), (window.original_height), false, false, 18)

		no_button:onClick(function(x, y) 
			PQS.handle_button_click(window.window_system_key, function()
				response_function(data, button_no_text)
				window:close()
			end)
		end)	
	end
	
	window:visible(true)
	
	PQS.window_system:add_window("confirm", window)	
	PQS.window_system:bring_to_front("confirm")
end

function PQS.handle_scale_click(scale, window_key)
	PQS.handle_button_click(window_key, function()
		PQS.load_player_settings()
		PQS.scale = scale
		PQS.player_settings[player.name .. "|access_button"]["scale"] = PQS.scale
		PQS.save_player_settings()
				
		PQS.window_system:bring_to_front(window_key)
		PQS.window_system:set_scale(PQS.scale)
		PQS.window_system:get_window(window_key):visible(true)
		PQS.window_system:get_window(window_key):move_to_center()
		PQS.window_system:bring_to_front(window_key)
		
		PQS.build_access_button()		
	end)
end

function PQS.handle_button_click(window_key, click_function)
	if (PQS.window_system.mouse_over_window_key == window_key) then
		PQS.window_system:bring_to_front(window_key)
	end
	if (PQS.window_system:get_window(window_key):is_top_window()) then
		click_function()
	end
end

function PQS.build_settings_window()
	if PQS.window_system:window_exists("settings") == true then 
		PQS.window_system:destroy_window("settings")
	end
	
	local window = PQS.PQS_WINDOW():new("Settings - PQS", 210, 210, PQS.scale)
		
	window.rebuild_content = function()
		local message_label = ui.label.add(0, 0, (window.original_width - 25), 50, "Scale - x" .. PQS.scale)
		window:add_element("message-label", message_label, 15, 45, false, false, 18)

		local button_x075 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x0.75</b>", "", 55, 30)
		window:add_element("scale-button-075", button_x075, ((window.original_width) - (58)), 42)

		button_x075:onClick(function(x, y) 
			PQS.handle_scale_click(.75, window.window_system_key)
		end)

		local button_x1 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x1</b>", "", 55, 30)
		window:add_element("scale-button-1", button_x1, ((window.original_width) - (58)), 75, false, false, 18)

		button_x1:onClick(function(x, y) 
			PQS.handle_scale_click(1, window.window_system_key)
		end)

		local button_x1_25 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x1.25</b>", "", 55, 30)
		window:add_element("scale-button-1-25", button_x1_25, ((window.original_width) - (58)), 108, false, false, 18)

		button_x1_25:onClick(function(x, y) 
			PQS.handle_scale_click(1.25, window.window_system_key)
		end)

		local button_x1_5 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x1.5</b>", "", 55, 30)
		window:add_element("scale-button-1-5", button_x1_5, ((window.original_width) - (58)), 141, false, false, 18)

		button_x1_5:onClick(function(x, y) 
			PQS.handle_scale_click(1.5, window.window_system_key)
		end)

		local button_x1_75 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x1.75</b>", "", 55, 30)
		window:add_element("scale-button-1-75", button_x1_75, ((window.original_width) - (58)), 174, false, false, 18)

		button_x1_75:onClick(function(x, y) 
			PQS.handle_scale_click(1.75, window.window_system_key)
		end)

		local button_x2 = ui.guiButton.add(0, 0, PQS.tiny_tex, "<b>x2</b>", "", 55, 30)
		window:add_element("scale-button-2", button_x2, ((window.original_width) - (58)), 207, false, false, 18)

		button_x2:onClick(function(x, y) 
			PQS.handle_scale_click(2, window.window_system_key)
		end)

		local reset_ui = ui.guiButton.add(0, 0, PQS.tiny_tex, "Reset PQS UI", "", 125, 30)
		window:add_element("reset-ui-button", reset_ui, 15, 207, false, false, 18)

		reset_ui:onClick(function(x, y) 
			PQS.handle_button_click(window.window_system_key, function()
				PQS.build_confirm_window("Reset Player Quest System UI", "This will reset all windows and buttons in PQS to their default positions. Are you sure you want to continue?", "Yes", "No", {quest = quest},
				function(data, response)
					if response == "Yes" then 
						PQS.handle_scale_click(1.25, "confirm")
						PQS.window_system:reset_window_positions()
						PQS.reset_access_button()
					end
				end)
			end)
		end)
		
	end
	
	window:visible(false)
	
	PQS.window_system:add_window("settings", window)	
end

function PQS.build_quests_window()
	
	if PQS.window_system:window_exists("quests") == true then 
		PQS.window_system:destroy_window("quests")
	end
	
	local window = PQS.PQS_WINDOW():new("Quests - PQS", 500, 300, PQS.scale)
	window.current_quest = 1
	
	window.rebuild_content = function()
		local quest_count = PQS.get_quest_count()	
		local quest = PQS.get_quest_by_index(window.current_quest)
	
		if quest ~= nil then 
			local enabled = (PQS.is_quest_enabled(quest:get_id()) == true)
			local color = "#00FF00"
			
			if (enabled == false) then
				color = "#FF0000"
			end
			
			local current_y = 50
			local spacing = 30
			
			local quest_title_text = "<b><i><color=" .. color .. ">" .. quest.quest_data['quest']['name'] .. "</color></i></b>"
			
			local label = ui.label.add(0, 0, (window.original_width - 25), 30, quest_title_text)
			window:add_element("quest-label", label, 15, current_y, false, false, 20)

			current_y = current_y + spacing

			local version_by_text = ""

			if (quest.quest_data['quest']['version'] ~= nil) then 
				version_by_text = version_by_text .. "v" .. quest.quest_data['quest']['version']
			end

			if (quest.quest_data['quest']['author'] ~= nil) then 
				if (version_by_text ~= "") then version_by_text = version_by_text .. " " end			
				version_by_text = version_by_text .. "<i>by</i> " .. quest.quest_data['quest']['author']
			end

			if (version_by_text ~= "") then 
				local version_by_label = ui.label.add(0, 0, (window.original_width - 25), 30, version_by_text)
				window:add_element("version-label", version_by_label, 20, current_y - 7, false, false, 14)
				current_y = current_y + spacing - 13
			end
			
			local backstory_label = ui.label.add(0, 0, (window.original_width - 25), 150, string.gsub(quest.quest_data['quest']['backstory'], "<br>", "\n"))
			window:add_element("backstory-label", backstory_label, 15, current_y, false, false, 18)
						
			if (window.current_quest < quest_count) then 
				local next_button = ui.guiButton.add(0, 0, PQS.tiny_tex, ">>", "", 30, 30)
				window:add_element("next-quest-button", next_button, ((window.original_width) - (42)), (window.original_height), false, false, 18)

				next_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						window.current_quest = window.current_quest + 1
						PQS.window_system:get_window(window.window_system_key):rebuild()
						PQS.window_system:get_window(window.window_system_key):visible(true)
					end)
				end)
			end
			if (window.current_quest > 1) then 
				local prev_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "<<", "", 30, 30)
				window:add_element("prev-quest-button", prev_button, 15, (window.original_height), false, false, 18)

				prev_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						window.current_quest = window.current_quest - 1
						PQS.window_system:get_window(window.window_system_key):rebuild()
						PQS.window_system:get_window(window.window_system_key):visible(true)
					end)
				end)
			end

			if (quest:get_journal_entry_count(player.name) > 0) then 

				local journal_button_text = "Journal"

				if (quest:get_first_unread_journal_entry_index(player.name) ~= nil) then
					journal_button_text = "<color=yellow>" .. journal_button_text .. "</color>"
				end
				
				local journal_button = ui.guiButton.add(0, 0, PQS.tiny_tex, journal_button_text, "", 100, 30)
				window:add_element("journal-button", journal_button, 50, (window.original_height), false, false, 18)

				journal_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						--PQS.build_dialog_window("Q1", "ABC")
						PQS.build_journal_window(quest)
						--PQS.window_system:get_window("journal"):visible(true)
						PQS.window_system:get_window(window.window_system_key):visible(false)
					end)
				end)			
			end
						
			if (quest:get_player_quest_item_list_count(player.name) > 0) then 
				local items_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Items", "", 100, 30)
				window:add_element("items-button", items_button, 150, (window.original_height), false, false, 18)

				items_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						PQS.build_items_window(quest)
						--PQS.build_dialog_window("Q1", "ABC")
						--PQS.window_system:get_window(window.window_system_key):visible(false)
					end)
				end)			
			end
			
			if (enabled) then 
				local deactivate_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Disable", "", 100, 30)
				window:add_element("deactivate-button", deactivate_button, 250, (window.original_height), false, false, 18)

				deactivate_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						local quest = PQS.get_quest_by_index(window.current_quest)
						if (quest ~= nil) then 
							PQS.disable_quest(quest:get_id())
							PQS.window_system:get_window(window.window_system_key):rebuild()
							PQS.window_system:get_window(window.window_system_key):visible(true)
						end
					end)
				end)			
			else 
				local activate_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Enable", "", 100, 30)
				window:add_element("activate-button", activate_button, 250, (window.original_height), false, false, 18)

				activate_button:onClick(function(x, y) 
					PQS.handle_button_click(window.window_system_key, function()
						local quest = PQS.get_quest_by_index(window.current_quest)
						if (quest ~= nil) then 
							PQS.enable_quest(quest:get_id())
							PQS.window_system:get_window(window.window_system_key):rebuild()
							PQS.window_system:get_window(window.window_system_key):visible(true)
						end
					end)
				end)			
			end
			
			local reset_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Reset", "", 100, 30)
			window:add_element("reset-button", reset_button, 350, (window.original_height), false, false, 18)

			reset_button:onClick(function(x, y) 
				PQS.handle_button_click(window.window_system_key, function()
					PQS.build_confirm_window("Reset Quest", "This will clear all quest progress. Are you sure you want to continue?", "Yes", "No", {quest = quest},
						function(data, response)
							if response == "Yes" then 
								data.quest:reset()
							end
						end
					)
				end)
			end)			
		end
	end
	
	window:visible(false)
	
	PQS.window_system:add_window("quests", window)
	
	return window
end

function PQS.build_journal_window(quest)
	
	if PQS.window_system:window_exists("journal") == true then 
		PQS.window_system:destroy_window("journal")
	end
	
	local window = PQS.PQS_WINDOW():new("Journal - PQS", 500, 300, PQS.scale)
	window.current_journal = quest:get_first_unread_journal_entry_index(player.name) or quest:get_journal_entry_count(player.name)
	window.current_quest = quest
	
	window.rebuild_content = function(self)
		--print(self.scale .. " | journal | " .. quest:get_id()) 
	
		local journal_entry_count = self.current_quest:get_journal_entry_count(player.name)	
		
		if (journal_entry_count < self.current_journal) then
			-- not available
		else 
		
			local journal_id = self.current_quest:get_journal_entry_by_index(player.name, window.current_journal)
			local journal = self.current_quest:get_journal(journal_id)
			
			--print(journal_id, journal)

			local color = "#69BBF7"
			
			if (self.current_quest:is_journal_entry_complete(player.name, journal_id)) then 
				color = "#00FF00"
			elseif (self.current_quest:is_journal_entry_fail(player.name, journal_id)) then
				color = "#FF0000"
			end
			
			local current_y = 50
			local spacing = 30
			
			local journal_title_text = "<b><i><color=" .. color .. ">" .. journal['title'] .. "</color></i></b>"
			
			local label = ui.label.add(0, 0, (self.original_width - 25), 30, journal_title_text)
			self:add_element("journal-title-label", label, 15, current_y, false, false, 20)

			current_y = current_y + spacing

			local journal_text_label = ui.label.add(0, 0, (self.original_width - 25), 150, string.gsub(journal['text'], "<br>", "\n"))
			self:add_element("journal-text-label", journal_text_label, 15, current_y, false, false, 18)

			if (journal["hint"] ~= nil and journal["hint"] ~= "") then 
				local hint_text = journal["hint"]
				hint_text = "<i>Hint: " .. hint_text .. "</i>"
				local hint_label = ui.label.add(0, 0, (self.original_width - 25), 30, hint_text)
				self:add_element("journal-hint-label", hint_label, 15, (self.original_height - 30), false, false, 18)
			end
			
		end
				
		-- nav buttons
		if (self.current_journal < journal_entry_count) then 
			local next_button = ui.guiButton.add(0, 0, PQS.tiny_tex, ">>", "", 30, 30)
			self:add_element("next-quest-button", next_button, ((self.original_width) - (42)), (self.original_height), false, false, 18)

			next_button:onClick(function(x, y) 
				PQS.handle_button_click(self.window_system_key, function()
					self.current_journal = self.current_journal + 1
					PQS.window_system:get_window(self.window_system_key):rebuild()
					PQS.window_system:get_window(self.window_system_key):visible(true)
				end)
			end)
		end
		if (self.current_journal > 1) then 
			local prev_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "<<", "", 30, 30)
			self:add_element("prev-quest-button", prev_button, 15, (self.original_height), false, false, 18)

			prev_button:onClick(function(x, y) 
				PQS.handle_button_click(self.window_system_key, function()
					self.current_journal = self.current_journal - 1
					PQS.window_system:get_window(self.window_system_key):rebuild()
					PQS.window_system:get_window(self.window_system_key):visible(true)
				end)
			end)
		end		
	end

	window.on_visible = function(self, is_visible)
		if (is_visible) then 
			local journal_id = self.current_quest:get_journal_entry_by_index(player.name, self.current_journal)
			if (self.current_quest:is_journal_entry_viewed(player.name, journal_id) == false) then 
				self.current_quest:set_journal_entry_viewed(player.name, journal_id)
				
				if (self.current_quest:journal_has_task(journal_id) == false) then
					-- mark it as complete upon view
					self.current_quest:set_journal_entry_complete(player.name, journal_id)
				end
				
				PQS.refresh_alert()
			end
		end
	end
	
	window:visible(true)
	PQS.window_system:add_window("journal", window)
	
	return window
end

function PQS.build_items_window(quest)
	
	if PQS.window_system:window_exists("items") == true then 
		PQS.window_system:destroy_window("items")
	end
	
	local window = PQS.PQS_WINDOW():new("Quest Items - PQS", 500, 300, PQS.scale)
	window.current_quest = quest
	
	window.rebuild_content = function(self)
	
		local quest_item_count = self.current_quest:get_player_quest_item_list_count(player.name)
		
		if (quest_item_count == 0) then
			-- none available
		else 
			local items = self.current_quest:get_player_quest_item_list(player.name)
			local color = "#69BBF7"
			local current_y = 50
			local spacing = 20
			local line_height = 20
			local characters_per_line = 80
		
			for item_id, item_quantity in pairs(items) do
				local item = self.current_quest:get_quest_item(item_id)
								
				local item_name_text = "<b><color=" .. color .. ">" .. item['name'] .. "</color></b>"
				
				if (item_quantity > 1) then
					item_name_text = item_name_text .. "<color=#FFA500> (" .. item_quantity .. ")</color>"
				end
				
				local label = ui.label.add(0, 0, (self.original_width - 25), 30, item_name_text)
				self:add_element("item-name-label-" .. item_id, label, 15, current_y, false, false, 16)

				if (item['description'] ~= nil) then 
					local line_count = math.ceil(#item['description'] / characters_per_line)				
					current_y = current_y + (spacing / 1.25)

					local description_label = ui.label.add(0, 0, (self.original_width - 25), (line_height * line_count), item['description'])
					self:add_element("item-description-name-label-" .. item_id, description_label, 15, current_y, false, false, 14)
					
					current_y = current_y + (spacing + (line_height * (line_count - 1)))
				else
					current_y = current_y + spacing
				end
			end
		end
	end

	window.on_visible = function(self, is_visible)
		if (is_visible) then 
		end
	end
	
	PQS.window_system:add_window("items", window)
	window:visible(true)
	PQS.window_system:bring_to_front("items")
	
	return window
end

function PQS.dialog_consolelog(source, consolelog_message, source_color, is_thought, message_color)

	is_thought = is_thought or false
	source_color = source_color or "FFA500"
	message_color = message_color or "FFFFFF"

	consolelog_message = PQS.substitute_text(consolelog_message)

	if is_thought then
		consolelog_message = "[i]" .. consolelog_message .. "[/i]"
	end

	ConsoleLog("[" .. source_color .. "]" .. source .. "[-]: [" .. message_color .. "]" .. consolelog_message .. "[-]")
end

function PQS.substitute_text(text, quest, dialog)
	quest = quest or nil
	dialog = dialog or nil

	local sub_text = text

	if (quest ~= nil) then 
		-- substitute any quest variables with their values
		if (quest.player_data[player_name] ~= nil) then
			for quest_variable_name, quest_variable_value in pairs (quest.player_data[player_name]) do
				sub_text = string.gsub(sub_text, "{" .. quest_variable_name .. "}", quest_variable_value)
			end
		end	
	end
	
	if (dialog ~= nil) then 
		sub_text = string.gsub(text, "{npc_name}", dialog.name)
	end
	
	sub_text = string.gsub(sub_text, "{player_name}", player.name)
	sub_text = string.gsub(sub_text, "<br>", "\n")
		
	-- TODO: allow pulling in player quest items, quest item data, location data, quest data, journal data
	
	return sub_text
end

function PQS.build_dialog_window(quest_id, dialog_id)

	if (dialog_id == false or quest_id == false) then return end
	
	-- player.name, player_text (optional), text (optional), dialog_topics (become buttons)
	-- inject messages into the message queue simulating an NPC conversation allowing to use the same event criteria interface
	
	if PQS.window_system:window_exists("dialog") == true then 
		PQS.window_system:destroy_window("dialog")
	end
	
	local window = PQS.PQS_WINDOW():new("Dialog - PQS", 400, 200, PQS.scale)
	window.current_dialog = dialog_id
	window.current_quest = quest_id
	window.message_list = {}
	window.messages_shown = false
	
	window.rebuild_content = function()
	
		local quest = PQS.get_quest(window.current_quest)				
		local dialog = quest.quest_data["dialog_" .. window.current_dialog]
		local dialog_topics = quest.quest_data["dialog_topics_" .. window.current_dialog] or nil
		
		-- 70 characters for the dialog width
		-- 20 pixels high per line
		
		if quest ~= nil and dialog ~= nil then 

			local current_y = 50
			local spacing_y = 25
			local line_count = 1
			local line_height = 25
			local characters_per_line = 50
			local topic_button_width = 100
			
			if (dialog.text_player_thought_pre ~= nil) then
				line_count = math.ceil(#dialog.text_player_thought_pre / characters_per_line)
				local text = PQS.substitute_text(dialog.text_player_thought_pre, quest, dialog)
				
				local text_player_thought_pre_label = ui.label.add(0, 0, (window.original_width - 25), (line_height * line_count), "<i>" .. text .. "</i>")
				window:add_element("text-player-thought-pre-label", text_player_thought_pre_label, 15, current_y, false, false, 16)
				
				window.message_list["text_player_thought_pre"] = {source = player.name, destination = player.name, message = text}
								
				current_y = current_y + (spacing_y + (line_height * (line_count - 1)))
			end 

			if (dialog.text_player_pre ~= nil) then
				line_count = math.ceil(#dialog.text_player_pre / characters_per_line)

				local text = PQS.substitute_text(dialog.text_player_pre, quest, dialog)
				
				local text_player_pre_label = ui.label.add(0, 0, (window.original_width - 25), (line_height * line_count), text)
				window:add_element("text-player-pre-label", text_player_pre_label, 15, current_y, false, false, 16)

				window.message_list["text_player_pre"] = {source = player.name, destination = dialog.name, message = text}
				
				current_y = current_y + (spacing_y + (line_height * (line_count - 1)))				
			end 

			if (dialog.text ~= nil) then
				local text_line = dialog.text
				local character_count = #text_line
				if dialog.name ~= nil then 
					character_count = character_count + #dialog.name + 2
					text_line = "<color=#FFA500>" .. dialog.name .. "</color>: " .. text_line
				end
				
				text_line = PQS.scale_text(PQS.substitute_text(text_line, quest, dialog), 16, PQS.scale)			
				line_count = math.ceil(character_count / characters_per_line)

				local text_label = ui.label.add(0, 0, (window.original_width - 25), (line_height * line_count), text_line)
				window:add_element("text-label", text_label, 15, current_y)

				window.message_list["text"] = {source = dialog.name, destination = player.name, message = dialog.text}
				
				current_y = current_y + (spacing_y + (line_height * (line_count - 1)))
			end 

			if (dialog.text_player_post ~= nil) then
				line_count = math.ceil(#dialog.text_player_post / characters_per_line)
				local text_player_post_label = ui.label.add(0, 0, (window.original_width - 25), (line_height * line_count), PQS.substitute_text(dialog.text_player_post, quest, dialog))
				window:add_element("text-player-post-label", text_player_post_label, 15, current_y, false, false, 16)

				window.message_list["text_player_post"] = {source = player.name, destination = dialog.name, message = dialog.text_player_post}
				
				current_y = current_y + (spacing_y + (line_height * (line_count - 1)))				
			end 

			if (dialog.text_player_thought_post ~= nil) then
				line_count = math.ceil(#dialog.text_player_thought_post / characters_per_line)
				
				local text_player_thought_post_label = ui.label.add(0, 0, (window.original_width - 25), (line_height * line_count), "<i>" .. PQS.substitute_text(dialog.text_player_thought_post, quest, dialog) .. "</i>")
				window:add_element("text-player-thought-post-label", text_player_thought_post_label, 15, current_y, false, false, 16)
				
				window.message_list["text_player_thought_post"] = {source = player.name, destination = player.name, message = dialog.text_player_thought_post}				
			end 
			
			local topics_bar = ui.texture.add(0, 0, "pqs/modules/window-background.png", false, 2, window.original_width, 35)
			window:add_element("topics-bar", topics_bar, 0, (window.original_height) + (spacing_y + 9))
			topics_bar:zIndex(window.zIndex)
		
			if dialog_topics ~= nil then 
				--local quest = PQS.get_quest(window.current_quest)
				local topic_count = 1
				local topic_x = 2
				
				for topic_key, topic_id in pairs (dialog_topics) do
								
					local dialog_topic = quest.quest_data["dialog_topic_" .. topic_id]
					local has_been_viewed = quest:is_dialog_viewed(player.name, dialog_topic.dialog)
					
					if (dialog_topic.show_if == nil or (quest:get_quest_variable(player.name, dialog_topic.show_if) == 1)) then
					
						local topic_label = dialog_topic.label
						
						if (dialog_topic.important ~= nil and dialog_topic.important == 1) then
							topic_label = "<color=#FFD700>" .. topic_label .. "</color>"
						end
						
						if (has_been_viewed == false) then
							topic_label = "<b>" .. topic_label .. "</b>"
						end
						
						local topic_button = ui.guiButton.add(0, 0, PQS.tiny_tex, topic_label, "", topic_button_width, 30)
						window:add_element("topic-" .. topic_key .. "-button", topic_button, topic_x, (window.original_height) + (spacing_y + 11), false, false, 14)

						topic_button:onClick(function(x, y) 
							PQS.handle_button_click(window.window_system_key, function()
								PQS.build_dialog_window(window.current_quest, dialog_topic.dialog)
							end)
						end)

						topic_count = topic_count + 1
						topic_x = (3) + (topic_button_width)
					end
				end
			end
			
			local topic_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Goodbye", "", 70, 30)
			window:add_element("topic-end-button", topic_button, ((window.original_width) - (72)), (window.original_height) + (spacing_y + 11), false, false, 14)

			topic_button:onClick(function(x, y) 
				PQS.handle_button_click(window.window_system_key, function()				
					PQS.dialog_consolelog(player.name, "Goodbye", "A39FCE")
					PQS.process_log_message('NPC', player.name, player.name .. " to " .. dialog.name .. " : " .. "Goodbye")
					window:close()
				end)
			end)
		end
		
		if (window.messages_shown == false and window.is_visible == true) then 
			for message_type, message_data in pairs(window.message_list) do
				local source_color = false
				local is_thought = (message_data.source == message_data.destination)
				
				if (message_data.source == player.name) then
					source_color = "A39FCE"
				end
			
				PQS.dialog_consolelog(message_data.source, message_data.message, source_color, is_thought)
				PQS.process_log_message('NPC', message_data.name, message_data.source .. " to " .. message_data.destination .. " : " .. message_data.message)
			end
			
			window.messages_shown = true
		end
	end
	
	window.on_visible = function(self, is_visible)
		if (is_visible) then		
			-- log the view of the topic
			PQS.get_quest(self.current_quest):set_dialog_viewed(player.name, self.current_dialog)		
		end
	end
	
	window:visible(false)
	
	PQS.window_system:add_window("dialog", window)
	
	window:visible(true)
	PQS.window_system:bring_to_front(window.window_system_key)
	
	return window
end

function PQS.build_menu()
	
	if PQS.window_system:window_exists("menu") == true then 
		PQS.window_system:destroy_window("menu")
	end
		
	local menu_window = PQS.PQS_WINDOW():new("Menu - PQS", 175, 110, PQS.scale)
		
	menu_window.rebuild_content = function()
			
		local menu_button_current = 50
		local button_left = (menu_window.original_width / 2) - (PQS.menu_button_width / 2)
		local button_height = 25
				
		local quests_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Quests", "", PQS.menu_button_width, button_height)
		menu_window:add_element("quests-button", quests_button, button_left, menu_button_current)
		menu_button_current = menu_button_current + PQS.menu_button_spacing

		quests_button:onClick(function(x, y) 
			PQS.handle_button_click(menu_window.window_system_key, function()
				PQS.window_system:get_window("menu"):visible(false)
				PQS.window_system:get_window("quests"):visible(true)
				PQS.window_system:bring_to_front("quests")
			end)
		end)
				
		local settings_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "Settings", "", PQS.menu_button_width, button_height)
		menu_window:add_element("settings-button", settings_button, button_left, menu_button_current)
		menu_button_current = menu_button_current + (PQS.menu_button_spacing)

		settings_button:onClick(function(x, y) 
			PQS.handle_button_click(menu_window.window_system_key, function()
				PQS.window_system:get_window("menu"):visible(false)
				PQS.window_system:get_window("settings"):visible(true)
				PQS.window_system:bring_to_front("settings")
			end)
		end)
	
		local about_button = ui.guiButton.add(0, 0, PQS.tiny_tex, "About", "", PQS.menu_button_width, button_height)
		menu_window:add_element("about-button", about_button, button_left, menu_button_current)
		menu_button_current = menu_button_current + PQS.menu_button_spacing

		about_button:onClick(function(x, y) 
			PQS.handle_button_click(menu_window.window_system_key, function()
				PQS.build_about_window()
			end)
		end)
		
	end
		
	menu_window:visible(false)
	
	PQS.window_system:add_window("menu", menu_window)	
end

function PQS.build_about_window()
	
	if PQS.window_system:window_exists("about") == true then 
		PQS.window_system:destroy_window("about")
	end
		
	local window = PQS.PQS_WINDOW():new("About - PQS", 400, 200, PQS.scale)
		
	window.rebuild_content = function(self)

		local label = ui.label.add(0, 0, (self.original_width - 25), 30, "The Player Quest System")
		self:add_element("about-title-label", label, 15, 50, false, false, 20)
		
		local version_label = ui.label.add(0, 0, (self.original_width - 25), 60, "v" .. PQS.version .. " created in 2021 by Garritt Grandberg \naka peace-n-love aka Tirrag aka Goriah Kuulas")
		window:add_element("version-label", version_label, 15, 80, false, false, 14)

		local thank_you_label = ui.label.add(0, 0, (self.original_width - 25), 60, "A huge thank you to Tristano Sforza whose sotaTales provided the inspiration and spark for this mod")
		window:add_element("thank-you-label", thank_you_label, 15, 120, false, false, 14)
	end
		
	window:visible(true)
	
	PQS.window_system:add_window("about", window)
	PQS.window_system:bring_to_front("about")	
end

-- ---------------------------------------------------------
-- Access button UI functions
-- ---------------------------------------------------------

function PQS.show_alert()
	PQS.alert_icon:visible(true)
end

function PQS.hide_alert()
	PQS.alert_icon:visible(false)
end

function PQS.is_alert_visible()
	return PQS.alert_icon:visible()
end

function PQS.refresh_alert()
	if (PQS.has_unread_journal_entry()) then 
		PQS.show_alert()
	else
		PQS.hide_alert()
	end
end

function PQS.has_unread_journal_entry()

	local has_unread = false

	for quest_id, quest in pairs(PQS.quests) do			
		if (PQS.is_quest_enabled(quest_id)) then 
			if (quest:get_first_unread_journal_entry_index(player.name) ~= nil) then 
				has_unread = true
				break
			end
		end
	end
	
	return has_unread
end

function PQS.reset_access_button()
	PQS.move_access_button(client.screen.width - (50 * PQS.scale), client.screen.height - (50 * PQS.scale), true)
end

function PQS.move_access_button(left, top, save)
	left = left or PQS.player_settings[player.name .. "|access_button"]['left']
	top = top or PQS.player_settings[player.name .. "|access_button"]['top']
	save = save or false
	
	PQS.access_button:moveTo(left, top)	

	PQS.player_settings[player.name .. "|access_button"]['left'] = left
	PQS.player_settings[player.name .. "|access_button"]['top'] = top	
	
	if (save) then
		PQS.load_player_settings()
		PQS.player_settings[player.name .. "|access_button"]['left'] = left
		PQS.player_settings[player.name .. "|access_button"]['top'] = top	
		PQS.save_player_settings()
	end
end

function PQS.build_access_button()

	local restore_alert_icon = PQS.is_alert_visible()

	PQS.load_player_settings()

	local prev_top = PQS.player_settings[player.name .. "|access_button"]['top']
	local prev_left = PQS.player_settings[player.name .. "|access_button"]['left']

	if (PQS.access_button ~= nil) then
		PQS.access_button:remove()
		PQS.access_button = nil
	end

	PQS.alert_icon:resizeTo((50 * PQS.scale), (50 * PQS.scale))
	PQS.alert_icon:moveTo(prev_left, prev_top)
	PQS.alert_icon:zIndex(10)

	PQS.access_button = ui.guiButton.add(prev_left, prev_top, PQS.tiny_tex, PQS.scale_text("PQS", 18, PQS.scale), "", (50 * PQS.scale), (50 * PQS.scale))
		
	PQS.access_button:onClick(function(x, y)
		if (ShroudGetKeyDown("LeftControl")) == false then 
			local window = PQS.window_system:get_window("menu")
			window:visible(window.is_visible == false)
			PQS.window_system:bring_to_front("menu")
		else
			PQS.access_button.is_moving = true
		end
	end)
	PQS.access_button.is_moving = false
	
	ui.onMouseMove(function(button_data, mouse_x, mouse_y)
		-- button_data, key is button number, value is down/held/up	
		if (PQS.access_button.is_moving == true) then
			PQS.access_button:moveTo(mouse_x, mouse_y)
			PQS.alert_icon:moveTo(mouse_x, mouse_y)
		end
	end)

	ui.onMouseButton(function(click_type, button_index, mouse_x, mouse_y)
		--"dblclicked/clicked/down/up"
		--print("mousebutton", click_type, button_index, mouse_x, mouse_y)
		--print(click_type, " | ", button_index)
		
	end)

	ui.onUpdate(function(click_type, button_index, mouse_x, mouse_y)
		if PQS.access_button.is_moving == true and ShroudGetKeyDown("LeftControl") == false then
			PQS.access_button.is_moving = false			
			PQS.move_access_button(PQS.access_button.rect.left, PQS.access_button.rect.top, true)			
		end
	end)

	-- is the button off the right edge
	if ((PQS.access_button.rect.left + PQS.access_button.rect.width) > client.screen.width) then
		local shift_amount = (PQS.access_button.rect.left + (PQS.access_button.rect.width)) - client.screen.width		
		PQS.move_access_button(PQS.access_button.rect.left - shift_amount, PQS.access_button.rect.top, true)
	end

	-- is the button off the bottom edge
	if ((PQS.access_button.rect.top + PQS.access_button.rect.height) > client.screen.height) then
		local shift_amount = (PQS.access_button.rect.top + (PQS.access_button.rect.height)) - client.screen.height
		PQS.move_access_button(PQS.access_button.rect.left, PQS.access_button.rect.top - shift_amount, true)
	end

	PQS.alert_icon:visible(restore_alert_icon)
	PQS.access_button:visible(true)
end

-- ---------------------------------------------------------
-- Utility functions
-- ---------------------------------------------------------

function PQS.get_table_element_by_number(table_to_search, target_element_number)
	local table_element = nil
	if (target_element_number == nil) then target_element_number = 1 end
	
	if (type(table_to_search) == 'table') then
		local current_element_number = 1
		for table_key, table_value in pairs(table_to_search) do
			if (current_element_number == target_element_number) then
				table_element = table_value
				break
			end
			
			current_element_number = current_element_number + 1
		end
	end
	
	return table_element
end

-- distance between two points
function PQS.distance(x1, y1, z1, x2, y2, z2)
	local distance = math.sqrt(((x2 - x1) ^ 2) + ((y2 - y1) ^ 2) + ((z2 - z1) ^ 2) * 1.0)				
	return distance
end

-- https://stackoverflow.com/questions/640642/how-do-you-copy-a-lua-table-by-value
function PQS.deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[PQS.deepcopy(orig_key)] = PQS.deepcopy(orig_value)
        end
        setmetatable(copy, PQS.deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

-- https://stackoverflow.com/questions/1426954/split-string-in-lua
--function PQS.split(s, delimiter)

    --result = {}
    --for match in (s..delimiter):gmatch("(.-)"..delimiter) do
    --    table.insert(result, match)
    --end
    --return result
--end

function PQS.clean_log_message(message)

	--local check_start, check_end = string.find(message,  "%: ")
	--if (check_start != nil) then
	--	message = string.sub(message, check_end + 1)
	--end

	local clean_message = message
					
	--ConsoleLog("stripping message: " .. message)
	local check_start, check_end = string.find(message,  "%[")
	if (check_start != nil) then
		local end_start, end_end = string.find(message, "%]", check_start + 1)
		if (end_start != nil) then
			clean_message = string.sub(message, 1, check_start - 1) .. string.sub(message, end_start + 1)
			clean_message = PQS.clean_log_message(clean_message)
		end
	end
									
	return clean_message
end

-- https://stackoverflow.com/questions/35006931/lua-line-breaks-in-strings
function PQS.formatUpToX(s, x, indent)
  x = x or 79
  indent = indent or ""
  local t = {""}
  local function cleanse(s) return s:gsub("@x%d%d%d",""):gsub("@r","") end
  for prefix, word, suffix, newline in s:gmatch("([ \t]*)(%S*)([ \t]*)(\n?)") do
    if #(cleanse(t[#t])) + #prefix + #cleanse(word) > x and #t > 0 then
      table.insert(t, word..suffix) -- add new element
    else -- add to the last element
      t[#t] = t[#t]..prefix..word..suffix
    end
    if #newline > 0 then table.insert(t, "") end
  end
  return indent..table.concat(t, "\n"..indent)
end

function PQS.process_log_message(type, src, msg)

	-- skip verbose combat
	if (string.find(msg,  ": Verbose Combat:")) then return end

	msg = PQS.clean_log_message(msg)

	if (type == 'CombatSelf') then
		local msg_split = PQS.split(msg, ' : ')
		local message = {message_type = "Combat", message = msg_split[2]}
		table.insert(PQS.console_log_message_queue, message)		
	elseif (type == 'CombatOthers') then -- party/pet
		local msg_split = PQS.split(msg, ' : ')
		local message = {message_type = "Combat", message = msg_split[2]}
		table.insert(PQS.console_log_message_queue, message)		
	elseif (type == 'NPC') then

		-- find the participants and individual message
		local msg_split = PQS.split(msg, ' : ')
		
		local to_split = PQS.split(msg_split[1], ' to ')

		local msg_to = to_split[2]
		local just_msg = msg_split[2]
	
		-- add the source if needed
		if (src == "" or src == nil) then
			src = to_split[1]
		end

		if (src ~= "") then
			-- src will be empty if npc is talking, will be player name when player
			local message = {message_type = type, source = src, destination = msg_to, message = just_msg}
			
			--print(message.message_type, " | ", message.source, " | ", message.destination, " | ", message.message)
			table.insert(PQS.console_log_message_queue, message)
		end
	elseif (type == 'Loot') and (
		string.find(msg, 'You picked up') ~= nil or 
		string.find(msg, 'You looted') ~= nil or 
		string.find(msg, 'You harvested') ~= nil or 
		string.find(msg, 'You gained') ~= nil or 
		string.find(msg, 'You won') ~= nil 
	) then -- harvest, loot, picked up, gained

		local msg_split = PQS.split(msg, ' : ')
		local just_msg = msg_split[2]

		-- test
		--just_msg = "You won Garnet (Unrefined Gemstone)!  It has been placed in your inventory."
		
		local item = nil
		local quantity = 1
		
		if (string.find(just_msg, 'You won')) then
			local item_split = PQS.split(just_msg, '!')		
			item = string.gsub(item_split[1], 'You won ', '')
			-- currently no quantity information.  could possibly obtain from other messages (value: )
		else
			--item = string.match(just_msg, "(%b())")
			item = string.match(just_msg, "%(([^%)]+)%)")
			
			-- pull out the quantity
			if (string.find(just_msg, " x") ~= nil) then
				local quantity_split = PQS.split(just_msg, " x")
				item = quantity_split[1]
				quantity = string.sub(quantity_split[2], 1, -2)
			end

			-- regex removes ending paren if there is one so add it back in
			if (item ~= nil and string.find(item, "%(") ~= nil) then 
				item = item .. ')'
			end
		end
						
		--print(item, ' -- ', quantity)
		
		local message = {message_type = type, item = item, quantity = quantity, message = just_msg}
		table.insert(PQS.console_log_message_queue, message)
	
	elseif (type == 'System') then -- skill gain, crafting
		if (string.find(msg, 'crafted') and string.find(msg, 'Recipe')) then
			local item_split = PQS.split(msg, 'Recipe: ')
			local item = item_split[2]:gsub("%.", ""):gsub("%)", "")
			
			local message = {message_type = "Craft", recipe = item}
			table.insert(PQS.console_log_message_queue, message)			
		end
	end
	
	--print(type, ' -- ', src, ' -- ', msg)
end

return PQS

--[[

-- test crafting messages

You crafted an exceptional item (Recipe: Copper Ingot).
Item crafted: Recipe: Copper Ingot.


-- test loot messages --

You picked up an item (Small Square Jar (White))
You picked up an item (Large Square Bottle)
You picked up 37 gold.
You looted an item (Kobold Slag Sword).
You harvested an item (Pine Wood).
You harvested an item (Obsidian Chip x26)
You harvested an item (Sapphire Fragment (Unrefined Gemstone) x6).
You gained an item (Young Mosasaurus (Lava Fish)) from fishing.
You gained an item (Coelacanth (Lava Fish)) from fishing.
You won Garnet (Unrefined Gemstone)!  It has been placed in your inventory.
You won Rusty Plate Gauntlets!  It has been placed in your inventory.
You won Supply Bundle (Grey)!  It has been placed in your inventory.
]]
