local PQS_DATA = {
	quest_filename = nil, 
	quest_data = nil,
	player_data = {},
	previous_player_data = {},
	last_player_data = {},
	LIP = nil
}

function PQS_DATA:new(quest_filename, LIP)
	setmetatable({}, PQS_DATA)
	
	self.quest_filename = quest_filename
	self.LIP = LIP

	if self:load_quest() == false then
		self = nil
	else
		self:load_player_data()
		self.last_player_data = self.LIP.load(ShroudLuaPath..'/pqs/' .. self.quest_data['quest']['id'] .. '.save')
	end

	return self
end

function PQS_DATA:reset()
	self.player_data[player.name] = {}
	self.player_data[player.name .. "|QuestItems"] = {}
	self.player_data[player.name .. "|DialogHistory"] = {}
	
	local journal_list = self:get_journal_list()
	
	for journal_key, journal_id in pairs(journal_list) do
		local task_list = self:get_task_list(player.name, journal_id)

		for task_id, task_status in pairs(task_list) do
			local task_item_list = self:get_task_item_list(player.name, journal_id)
			self:clear_task_items(player.name, journal_id, task_id)
		end
		
		self:clear_task_list(player.name, journal_id)
	end
	
	self.player_data[player.name .. "|Journal"] = {}
	self:clear_timers(player.name)
	
	self:save_user_settings()
end

function PQS_DATA:get_id()
	return self.quest_data['quest']['id']
end

function PQS_DATA:get_start_event_id()
	local start_event_id = nil
	
	if (self.quest_data['quest']['start_event'] ~= nil) then
		start_event_id = self.quest_data['quest']['start_event']
	end
	
	return start_event_id
end

function PQS_DATA:load_player_data()
	
	local quest_id = self.quest_data['quest']['id']

	-- ensure the user setting file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath..'/pqs/' .. quest_id .. '.save', "r")
	if file == nil then 
		file = io.open(ShroudLuaPath..'/pqs/' .. quest_id .. '.save', "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end
	
	self.player_data = self.LIP.load(ShroudLuaPath..'/pqs/' .. quest_id .. '.save')	
	self.previous_player_data = self.LIP.load(ShroudLuaPath..'/pqs/' .. quest_id .. '.save')

end

function PQS_DATA:save_user_settings()
	local quest_id = self.quest_data['quest']['id']
	self.LIP.save(ShroudLuaPath..'/pqs/' .. quest_id .. '.save', self.player_data)
end

function PQS_DATA:load_quest()

	local quest_loaded = false

	if self:quest_file_exists() then
		self.quest_data = self.LIP.load(ShroudLuaPath..'/pqs/quests/' .. self.quest_filename)		
		quest_loaded = true
	end
	
	return quest_loaded
end

function PQS_DATA:dump()
	for setting_key, setting_value in pairs(self.quest_data) do 
		print("--", setting_key, "--")
		for subsetting_key, subsetting_value in pairs(self.quest_data[setting_key]) do
			print(subsetting_key, " : ", subsetting_value)
		end
	end	
end

function PQS_DATA:quest_file_exists()
	local file_exists = false

	-- make sure the quest settings file exists
	local file = io.open(ShroudLuaPath..'/pqs/quests/' .. self.quest_filename, "r")
	if file then 
		file_exists = true
		file:close() 
	else
		ConsoleLog("[FF0000]Unable to locate quest settings file " .. self.quest_filename .. "[-]")
	end
	
	return file_exists
end

-- ---------------------------------------------------------
-- Quest variable functions
-- ---------------------------------------------------------

function PQS_DATA:set_quest_variable(player_name, variable_name, variable_value)
	
	local quest_id = self.quest_data['quest']['id']
	
	if (self.player_data[player_name] == nil) then
		self.player_data[player_name] = {}
	end
	
	if (self.player_data[player_name] ~= nil and self.player_data[player_name][variable_name] ~= nil) then
		if (self.previous_player_data[player_name] == nil) then self.previous_player_data[player_name] = {} end
		self.previous_player_data[player_name][variable_name] = self.player_data[player_name][variable_name]
	end
	
	self.player_data[player_name][variable_name] = variable_value
	
	self:save_user_settings()
end

function PQS_DATA:is_quest_variable(player_name, variable_name)
	return (self.player_data[player_name] ~= nil and self.player_data[player_name][variable_name] ~= nil)
end

function PQS_DATA:get_quest_variable(player_name, variable_name, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local quest_id = self.quest_data['quest']['id']
	local quest_variable_value = nil
	
	if (use_previous == true) then
		if (self.previous_player_data[player_name] ~= nil and self.previous_player_data[player_name][variable_name] ~= nil) then
			quest_variable_value = self.previous_player_data[player_name][variable_name]
		end
	elseif (use_last == true) then
		if (self.last_player_data[player_name] ~= nil and self.last_player_data[player_name][variable_name] ~= nil) then
			quest_variable_value = self.last_player_data[player_name][variable_name]
		end
	else 
		if (self.player_data[player_name] ~= nil and self.player_data[player_name][variable_name] ~= nil) then
			quest_variable_value = self.player_data[player_name][variable_name]
		end
	end	
	
	return quest_variable_value
end

function PQS_DATA:set_last_player_data()
	self.last_player_data = {}
	for key, value in pairs(self.player_data) do
		self.last_player_data[key] = {}
		for subkey, subvalue in pairs(value) do
			self.last_player_data[key][subkey] = subvalue
		end
	end
end

-- ---------------------------------------------------------
-- Player quest item functions
-- ---------------------------------------------------------

function PQS_DATA:set_player_quest_item(player_name, item_id, value)
	local inventory_name = player_name .. "|QuestItems"	
	self:set_quest_variable(inventory_name, item_id, value)
end

function PQS_DATA:add_player_quest_item(player_name, item_id, quantity)
	if (quantity == nil) then quantity = 1 end
	
	local current_quantity = self:get_player_quest_item_count(player_name, item_id)	
	current_quantity = current_quantity + quantity
	
	self:set_player_quest_item(player_name, item_id, current_quantity)
	
	return current_quantity
end

function PQS_DATA:remove_player_quest_item(player_name, item_id, quantity)
	if (quantity == nil) then quantity = 1 end

	local current_quantity = self:get_player_quest_item_count(player_name, item_id)	

	if (current_quantity >= quantity) then
		current_quantity = current_quantity - quantity
	else
		current_quantity = 0
	end
	
	self:set_quest_item(player_name, item_id, current_quantity)
	
	return current_quantity
end

function PQS_DATA:clear_player_quest_items(player_name)
	local inventory_name = player_name .. "|QuestItems"
	
	self.player_data[inventory_name] = {}
	self.previous_player_data[inventory_name] = {}
	
	self:save_user_settings()
end

function PQS_DATA:get_player_quest_item_count(player_name, item_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false

	local inventory_name = player_name .. "|QuestItems"

	local current_quantity = self:get_quest_variable(inventory_name, item_id, use_previous, use_last)
	if (current_quantity == nil) then current_quantity = 0 end
	
	return current_quantity
end

function PQS_DATA:get_player_quest_item_list(player_name)
	local inventory_name = player_name .. "|QuestItems"
	local quest_item_list = {}

	if (self.player_data[inventory_name] ~= nil) then		
		quest_item_list = self.player_data[inventory_name]		
	end
	
	return quest_item_list
end

function PQS_DATA:get_player_quest_item_list_count(player_name, count_all)
	count_all = count_all or false
	local inventory_name = player_name .. "|QuestItems"
	local quest_item_list_count = 0

	if (self.player_data[inventory_name] ~= nil) then		
		for item_id, item_quantity in pairs(self.player_data[inventory_name]) do
			if (count_all) then 
				quest_item_list_count = quest_item_list_count + item_quantity
			else
				quest_item_list_count = quest_item_list_count + 1
			end
		end
	end
	
	return quest_item_list_count
end

-- ---------------------------------------------------------
-- Dialog functions
-- ---------------------------------------------------------

function PQS_DATA:set_dialog_viewed(player_name, dialog_id)
	local dialog_history_name = player_name .. "|DialogHistory"	
	self:set_quest_variable(dialog_history_name, dialog_id, 1)
end

function PQS_DATA:clear_dialog_viewed(player_name, dialog_id)
	local dialog_history_name = player_name .. "|DialogHistory"
	self:set_quest_variable(dialog_history_name, dialog_id, "")
end

function PQS_DATA:is_dialog_viewed(player_name, dialog_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local dialog_history_name = player_name .. "|DialogHistory"
	local viewed = self:get_quest_variable(dialog_history_name, dialog_id, use_previous, use_last)
	return viewed ~= nil and viewed == 1
end

function PQS_DATA:clear_dialogs(player_name)
	local dialog_history_name = player_name .. "|DialogHistory"
	
	self.player_data[dialog_history_name] = {}
	self.previous_player_data[dialog_history_name] = {}
	
	self:save_user_settings()
end

function PQS_DATA:get_dialog_viewed_list(player_name)
	local dialog_history_name = player_name .. "|DialogHistory"
	local dialog_viewed_list = {}

	if (self.player_data[dialog_history_name] ~= nil) then
		dialog_viewed_list = self.player_data[dialog_history_name]
	end
	
	return dialog_viewed_list
end

-- ---------------------------------------------------------
-- Timer functions
-- ---------------------------------------------------------

function PQS_DATA:create_timer(player_name, timer_id)
	local timer_block_name = player_name .. "|Timers"	
	self:set_quest_variable(timer_block_name, timer_id, 0)
	self:set_quest_variable(player_name, timer_id, 0)
end

function PQS_DATA:destroy_timer(player_name, timer_id)
	local timer_block_name = player_name .. "|Timers"	
	self:set_quest_variable(timer_block_name, timer_id, "")
	self:set_quest_variable(player_name, timer_id, "")
end

function PQS_DATA:reset_timer(player_name, timer_id)
	local timer_block_name = player_name .. "|Timers"	
	self:set_quest_variable(timer_block_name, timer_id, 0)
	self:set_quest_variable(player_name, timer_id, 0)
end

function PQS_DATA:start_timer(player_name, timer_id)
	self:set_timer_status(player_name, timer_id, 1)
end

function PQS_DATA:stop_timer(player_name, timer_id)
	self:set_timer_status(player_name, timer_id, 2)
end

function PQS_DATA:update_timer_ticks(player_name, timer_id, ticks)
	ticks = ticks or 1
	local current_ticks = self:get_quest_variable(player_name, timer_id)
	current_ticks = current_ticks or 0

	self:set_quest_variable(player_name, timer_id, current_ticks + ticks)
end

function PQS_DATA:get_timer_ticks(player_name, timer_id)
	local current_ticks = self:get_quest_variable(player_name, timer_id)
	return current_ticks or 0
end

function PQS_DATA:get_timer_status(player_name, timer_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local timer_block_name = player_name .. "|Timers"	
	local status = self:get_quest_variable(timer_block_name, timer_id, use_previous, use_last)
	return status or 0
end

function PQS_DATA:set_timer_status(player_name, timer_id, status)
	local timer_block_name = player_name .. "|Timers"	
	self:set_quest_variable(timer_block_name, timer_id, status)
end

function PQS_DATA:clear_timers(player_name)
	local timer_block_name = player_name .. "|Timers"	
	
	self.player_data[timer_block_name] = {}
	self.previous_player_data[timer_block_name] = {}
	
	self:save_user_settings()
end

function PQS_DATA:get_timer_list(player_name)
	local timer_block_name = player_name .. "|Timers"	
	local timer_list = {}

	if (self.player_data[timer_block_name] ~= nil) then
		timer_list = self.player_data[timer_block_name]
	end
	
	return timer_list
end

function PQS_DATA:get_active_timer_list(player_name)
	local active_timer_list = {}
	local timer_list = self:get_timer_list(player_name)
	
	for timer_id, timer_status in pairs(timer_list) do
		if (timer_status == 1) then
			table.insert(active_timer_list, timer_id)
		end
	end
	
	return active_timer_list
end

function PQS_DATA:is_timer(player_name, timer_id)
	local timer_block_name = player_name .. "|Timers"
	return self:is_quest_variable(timer_block_name, timer_id)
end

-- ---------------------------------------------------------
-- Journal functions
-- ---------------------------------------------------------

function PQS_DATA:get_journal_list()
	local journal_entry_list = {}
	if (self.quest_data['journal_entries'] ~= nil) then
		journal_entry_list = self.quest_data['journal_entries']
	end
	return journal_entry_list
end

function PQS_DATA:get_journal(journal_id)
	local journal_entry = nil
	
	if (self.quest_data['journal_entry_' .. journal_id] ~= nil) then
		journal_entry = self.quest_data['journal_entry_' .. journal_id]
	end
	
	return journal_entry
end

function PQS_DATA:journal_has_task(journal_id)
	local journal_entry = nil
	local has_task = false
	
	if (self.quest_data['journal_entry_' .. journal_id] ~= nil) then
		journal_entry = self.quest_data['journal_entry_' .. journal_id]
		
		if (journal_entry ~= nil) then
			if (journal_entry["has_task"] ~= nil and journal_entry["has_task"] > 0) then
				has_task = true
			end
		end
	end
	
	return has_task
end

function PQS_DATA:get_journal_task_list(journal_id)
	local journal_task_list = {}
	if (self.quest_data['journal_tasks_' .. journal_id] ~= nil) then
		journal_task_list = self.quest_data['journal_tasks_' .. journal_id]
	end
	return journal_task_list
end

-- ---------------------------------------------------------
-- Player journal entry functions
-- ---------------------------------------------------------

function PQS_DATA:add_journal_entry(player_name, journal_id)

	self:clear_journal_entry_viewed(player_name, journal_id)
	
	-- add tasks for the journal entry
	local task_list = self:get_journal_task_list(journal_id)
	
	for task_key, task_id in pairs(task_list) do
		self:add_task(player_name, journal_id, task_id)
	end
end

function PQS_DATA:has_journal_entry(player_name, journal_id)
	local journal_name = player_name .. "|Journal"
	return (self:get_quest_variable(journal_name, journal_id) ~= nil)
end

function PQS_DATA:set_journal_entry_status(player_name, journal_id, status)
	local journal_name = player_name .. "|Journal"
	self:set_quest_variable(journal_name, journal_id, status)
end

function PQS_DATA:set_journal_entry_complete(player_name, journal_id)
	self:set_journal_entry_status(player_name, journal_id, 10)
end

function PQS_DATA:set_journal_entry_fail(player_name, journal_id)
	self:set_journal_entry_status(player_name, journal_id, 5)
end

function PQS_DATA:set_journal_entry_viewed(player_name, journal_id)
	self:set_journal_entry_status(player_name, journal_id, 1)
end

function PQS_DATA:clear_journal_entry_viewed(player_name, journal_id)
	self:set_journal_entry_status(player_name, journal_id, 0)
end

function PQS_DATA:is_journal_entry_viewed(player_name, journal_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local journal_name = player_name .. "|Journal"
	local journal_status = self:get_quest_variable(journal_name, journal_id, use_previous, use_last)
	return journal_status ~= nil and journal_status >= 1
end

function PQS_DATA:is_journal_entry_complete(player_name, journal_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local journal_name = player_name .. "|Journal"
	local journal_status = self:get_quest_variable(journal_name, journal_id, use_previous, use_last)
	return journal_status ~= nil and journal_status >= 10
end

function PQS_DATA:is_journal_entry_fail(player_name, journal_id, use_previous, use_last)
	use_previous = use_previous or false
	use_last = use_last or false
	local journal_name = player_name .. "|Journal"
	local journal_status = self:get_quest_variable(journal_name, journal_id, use_previous, use_last)
	return journal_status ~= nil and journal_status >= 5 and journal_status < 10
end

function PQS_DATA:clear_journal_entry(player_name)
	local journal_name = player_name .. "|Journal"
	
	self.player_data[journal_name] = {}
	self.previous_player_data[journal_name] = {}
	
	self:save_user_settings()
end

function PQS_DATA:get_journal_entry_list(player_name)
	local journal_name = player_name .. "|Journal"
	local journal_list = {}

	if (self.player_data[journal_name] ~= nil) then
		journal_list = self.player_data[journal_name]
	end
	
	return journal_list
end

function PQS_DATA:get_journal_entry_count(player_name)
	local count = 0
	local journal_entry_list = self:get_journal_entry_list(player_name)
	
	for journal_id, viewed in pairs(journal_entry_list) do
		count = count + 1
	end
	return count
end

function PQS_DATA:get_journal_entry_by_index(player_name, journal_entry_index)
	local current_index = 1
	local found_journal_id = nil
	local journal_entry_list = self:get_journal_entry_list(player_name)
	
	for journal_id, viewed in pairs(journal_entry_list) do
		if current_index == journal_entry_index then 
			found_journal_id = journal_id
			break
		end
		
		current_index = current_index + 1
	end
	
	return found_journal_id
end

function PQS_DATA:get_first_unread_journal_entry_index(player_name)
	local current_index = 1
	local found_index = nil
	local journal_entry_list = self:get_journal_entry_list(player_name)
	
	for journal_id, viewed in pairs(journal_entry_list) do
		if viewed == 0 then 
			found_index = current_index
			break
		end
		
		current_index = current_index + 1
	end
	
	return found_index
end


-- ---------------------------------------------------------
-- Task functions
-- ---------------------------------------------------------

function PQS_DATA:add_task(player_name, journal_id, task_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	self:set_quest_variable(journal_task_name, task_id, 0)
end

function PQS_DATA:get_task_list(player_name, journal_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	local journal_task_list = {}

	if (self.player_data[journal_task_name] ~= nil) then
		journal_task_list = self.player_data[journal_task_name]
	end
	
	return journal_task_list
end

function PQS_DATA:set_task_complete(player_name, journal_id, task_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	self:set_quest_variable(journal_task_name, task_id, 1)
end

function PQS_DATA:is_task_complete(player_name, journal_id, task_id)
	use_previous = use_previous or false
	use_last = use_last or false
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	local complete = self:get_quest_variable(journal_task_name, task_id, use_previous, use_last)
	return complete ~= nil and complete == 1
end

function PQS_DATA:set_task_failed(player_name, journal_id, task_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	self:set_quest_variable(journal_task_name, task_id, 2)
end

function PQS_DATA:is_task_failed(player_name, journal_id, task_id)
	use_previous = use_previous or false
	use_last = use_last or false
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	local complete = self:get_quest_variable(journal_task_name, task_id, use_previous, use_last)
	return complete ~= nil and complete == 2
end

function PQS_DATA:reset_task(player_name, journal_id, task_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	self:set_quest_variable(journal_task_name, task_id, 0)
end

function PQS_DATA:clear_task_list(player_name, journal_id)
	local journal_task_name = player_name .. "|" .. journal_id .. "|JournalTasks"
	
	self.player_data[journal_task_name] = {}
	self.previous_player_data[journal_task_name] = {}
	
	self:save_user_settings()
end

-- ---------------------------------------------------------
-- Player task item functions
-- ---------------------------------------------------------

function PQS_DATA:set_task_item(player_name, journal_id, task_id, item_id, value)
	local inventory_name = player_name .. "|" .. journal_id .. "|" .. task_id .. "|QuestItems"	
	self:set_quest_variable(inventory_name, item_id, value)
end

function PQS_DATA:add_task_item(player_name, journal_id, task_id, item_id, quantity)
	quantity = quantity or 1
	
	local current_quantity = self:get_task_item(player_name, journal_id, task_id, item_id)	
	current_quantity = current_quantity + quantity
	
	self:set_task_item(player_name, journal_id, task_id, item_id, current_quantity)
	
	return current_quantity
end

function PQS_DATA:remove_task_item(player_name, journal_id, task_id, item_id, quantity)
	quantity = quantity or 1

	local current_quantity = self:get_task_item_count(player_name, journal_id, task_id, item_id)

	if (current_quantity >= quantity) then
		current_quantity = current_quantity - quantity
	else
		current_quantity = 0
	end
	
	self:set_task_item(player_name, journal_id, task_id, item_id, current_quantity)
	
	return current_quantity
end

function PQS_DATA:clear_task_items(player_name, journal_id, task_id)
	local inventory_name = player_name .. "|" .. journal_id .. "|" .. task_id .. "|QuestItems"	
	
	self.player_data[inventory_name] = {}
	self.previous_player_data[inventory_name] = {}
	
	self:save_user_settings()
end

function PQS_DATA:get_task_item_count(player_name, journal_id, task_id, item_id)
	local inventory_name = player_name .. "|" .. journal_id .. "|" .. task_id .. "|QuestItems"	

	local current_quantity = self:get_quest_variable(inventory_name, item_id)
	if (current_quantity == nil) then current_quantity = 0 end
	
	return current_quantity
end

function PQS_DATA:get_task_item_list(player_name, journal_id, task_id)
	local inventory_name = player_name .. "|" .. journal_id .. "|" .. task_id .. "|QuestItems"	
	local quest_item_list = {}

	if (self.player_data[inventory_name] ~= nil) then
		quest_item_list = self.player_data[inventory_name]
	end
	
	return quest_item_list
end

-- ---------------------------------------------------------
-- Location functions
-- ---------------------------------------------------------

function PQS_DATA:get_location_list()
	local location_list = {}
		
	if (self.quest_data['locations'] ~= nil) then
		location_list = self.quest_data['locations']
	end
	
	return location_list
end

function PQS_DATA:get_location(location_id)
	local location = nil

	if (self.quest_data['location_' .. location_id] ~= nil) then
		location = self.quest_data['location_' .. location_id]
	end
	
	return location
end

-- ---------------------------------------------------------
-- Quest item functions
-- ---------------------------------------------------------

function PQS_DATA:get_quest_item_list()
	local quest_item_list = {}
	if (self.quest_data['quest_items'] ~= nil) then
		quest_item_list = self.quest_data['quest_items']
	end
	return quest_item_list
end

function PQS_DATA:get_quest_item(item_id)
	local quest_item = nil
	
	if (self.quest_data['item_' .. item_id] ~= nil) then
		quest_item = self.quest_data['item_' .. item_id]
	end
	
	return quest_item
end

-- ---------------------------------------------------------
-- Event functions
-- ---------------------------------------------------------

function PQS_DATA:get_event_list()
	local event_list = {}
	if (self.quest_data['events'] ~= nil) then
		event_list = self.quest_data['events']
	end
	return event_list
end

function PQS_DATA:get_event(event_id)
	local event = nil
	
	if (self.quest_data['event_' .. event_id] ~= nil) then
		event = self.quest_data['event_' .. event_id]
	end
	
	return event
end

function PQS_DATA:get_event_criteria_list(event_id)
	local event_criteria_list = {}

	if (self.quest_data['event_criteria_' .. event_id] ~= nil) then
		event_criteria_list = self.quest_data['event_criteria_' .. event_id]
	end
	
	return event_criteria_list
end

function PQS_DATA:get_event_criteria(event_id, criteria_id)
	local event_criteria = nil

	if (self.quest_data['event_criteria_' .. event_id .. '_' .. criteria_id] ~= nil) then
		event_criteria = self.quest_data['event_criteria_' .. event_id .. '_' .. criteria_id]
	end
	
	return event_criteria
end

function PQS_DATA:get_event_action_list(event_id)
	local event_action_list = {}
	
	if (self.quest_data['event_actions_' .. event_id] ~= nil) then
		event_action_list = self.quest_data['event_actions_' .. event_id]
	end
	
	return event_action_list
end

function PQS_DATA:get_event_action(event_id, action_id)
	local event_action = nil
	
	if (self.quest_data['event_action_' .. event_id .. '_' .. action_id] ~= nil) then
		event_action = self.quest_data['event_action_' .. event_id .. '_' .. action_id]
	end
	
	return event_action
end

return PQS_DATA