
local PQS_WINDOW = {
	LIP = nil,
	window_system = nil,
	window_system_key = nil,
	settings_file = "user.ini",
	player_settings = {},
	elements = {},
	left = 0,
	top = 0,
	width = 0,
	height = 0,
	original_width = 0,
	original_height = 0,
	title = "",
	is_visible = false,
	is_moving = false,
	is_mouse_over_window = false,
	last_is_mouse_over_window = false,
	mouse_over_element = {},
	last_mouse_over_element = {},
	scale = 1.0,
	title_tex = nil,
	body_tex = nil,
	rebuild_content = nil,
	on_mouse_move = nil,
	on_mouse_button = nil,
	on_mouse_enter = nil,
	on_mouse_exit = nil,
	on_mouse_move_element = nil,
	on_mouse_button_element = nil,
	on_mouse_enter_element = nil,
	on_mouse_exit_element = nil,
	on_visible = nil,
	zIndex = 1,
	controls_zIndex = 15,
	hide_move_button = false,
	hide_close_button = false
}

function PQS_WINDOW:track_mouse_for_elements()
	return self.on_mouse_button_element ~= nil or 
		self.on_mouse_enter_element ~= nil or 
		self.on_mouse_exit_element ~= nil or 
		self.on_mouse_move_element ~= nil 
end

function PQS_WINDOW:is_top_window()
	return (self.zIndex == 100)
end

function PQS_WINDOW:is_mouse_over(mouse_x, mouse_y)
	local is_mouse_over = false

	is_mouse_over = self.is_visible and (mouse_x >= self.left and mouse_y >= self.top) and 
		(mouse_x <= self.left + self.width and mouse_y <= self.top + (self.height + 30)) 
	
	--print(self.title, " | ", mouse_x, ", ", mouse_y, " | ", self.left, ", ", self.top, " | ", self.left + (self.width * self.scale), ", ", self.top + (self.height * self.scale), " | ", is_mouse_over)
	--print(self.scale, " | ", self.title, " | ", mouse_x, ", ", mouse_y, " | ", self.left, ", ", self.top, " | ", self.left + self.width, ", ", self.top + self.height, " | ", is_mouse_over)
	
	return is_mouse_over
end

function PQS_WINDOW:is_mouse_over_element(element_name, mouse_x, mouse_y)
	local is_mouse_over = false

	if (self.is_visible and self.is_mouse_over_window and self.elements[element_name] ~= nil) then 
		local element = self.elements[element_name].element
		
		is_mouse_over = self.is_visible and (mouse_x >= element.rect.left and mouse_y >= element.rect.top) and 
			(mouse_x <= element.rect.left + element.rect.width and mouse_y <= element.rect.top + element.rect.height) 
	end 
	
	--print(self.title, " | ", mouse_x, ", ", mouse_y, " | ", self.left, ", ", self.top, " | ", self.left + (self.width * self.scale), ", ", self.top + (self.height * self.scale), " | ", is_mouse_over)
	
	return is_mouse_over
end

function PQS_WINDOW:new(title, width, height, scale, left, top, hide_move_button, hide_close_button)
	setmetatable({}, PQS_WINDOW)
		
	self.LIP = ui.module.add("pqs/modules/LIP.lua")

	self.title = title or ""
	self.scale = scale or 1.0
	self.width = width or 0
	self.height = height or 0
	
	self.original_width = self.width
	self.original_height = self.height
	
	self.width = self.width * self.scale
	self.height = self.height * self.scale
	
	self.hide_move_button = hide_move_button or false
	self.hide_close_button = hide_close_button or false
	
	-- set to center of the screen
	local center_left = ((client.screen.width / 2) - (self.width / 2))
	local center_top = ((client.screen.height / 2) - (self.height / 2))

	self.left = left or center_left
	self.top = top or center_top
		
	self:load_player_settings()

	ui.onUpdate(function()

		self.is_mouse_over_window = self:is_mouse_over(client.mouse.x, client.mouse.y) and 
			(self.window_system:get_window_key_under_mouse(client.mouse.x, client.mouse.y) == self.window_system_key)
			
		if (self.is_mouse_over_window == true and self.last_is_mouse_over_window == false) then
			if (self.on_mouse_enter ~= nil) then
				self.on_mouse_enter()
			end
		elseif (self.is_mouse_over_window == false and self.last_is_mouse_over_window == true) then
			if (self.on_mouse_exit ~= nil) then
				self.on_mouse_exit()
			end
		end
		self.last_is_mouse_over_window = self.is_mouse_over_window
	end)

	ui.onMouseMove(function(button_data, mouse_x, mouse_y)
	
		-- button_data, key is button number, value is down/held/up	
		if (self.is_moving == true) then
			self:moveTo(mouse_x - (self.elements['move-button'].relative_left * self.scale), mouse_y)
		end
		
		if (self.is_mouse_over_window) then
			if (self.on_mouse_move ~= nil) then 
				self.on_mouse_move(button_data, mouse_x, mouse_y)
			end
			
			if self:track_mouse_for_elements() then			
				for element_name, element in pairs(self.elements) do
					self.last_mouse_over_element[element_name] = self.mouse_over_element[element_name] or false
					
					self.mouse_over_element[element_name] = self:is_mouse_over_element(element_name, mouse_x, mouse_y)
					
					if (self.mouse_over_element[element_name] and self.on_mouse_move_element ~= nil) then 
						self.on_mouse_move_element(element_name, button_data, mouse_x, mouse_y)
					end
					
					if (self.last_mouse_over_element[element_name] == false and self.mouse_over_element[element_name] == true and self.on_mouse_enter_element ~= nil) then 
						self.on_mouse_enter_element(element_name, button_data, mouse_x, mouse_y)
					elseif (self.last_mouse_over_element[element_name] == true and self.mouse_over_element[element_name] == false and self.on_mouse_exit_element ~= nil) then 
						self.on_mouse_exit_element(element_name, button_data, mouse_x, mouse_y)
					end
				end
			end
		end
		
	end)

	ui.onMouseButton(function(click_type, button_index, mouse_x, mouse_y)
		--"dblclicked/clicked/down/up"
		--print("mousebutton", click_type, button_index, mouse_x, mouse_y)
		--print(click_type, " | ", button_index)
		
		if (click_type == "up" and self.is_moving == true) then 
			self.is_moving = false
			self:load_player_settings(true)
			self.player_settings[player.name .. "|" .. self.title]['left'] = self.left
			self.player_settings[player.name .. "|" .. self.title]['top'] = self.top
			self:save_player_settings()			
		end		
		
		if (self.is_mouse_over_window) then
			if (self.on_mouse_button ~= nil) then 
				self.on_mouse_button(click_type, button_index, mouse_x, mouse_y)
			end

			if self.on_mouse_button_element ~= nil then 
				for element_name, element in pairs(self.elements) do
					if self.mouse_over_element[element_name] ~= nil and self.mouse_over_element[element_name] then 
						self.on_mouse_button_element(element_name, click_type, button_index, mouse_x, mouse_y)
					end 				
					--print(element_name, " | ", self.mouse_over_element[element_name])
				end
			end
		end
		
	end)
	
	return self
end

function PQS_WINDOW:move_to_center()
	local center_left = ((client.screen.width / 2) - (self.width / 2))
	local center_top = ((client.screen.height / 2) - (self.height / 2))
	
	self:moveTo(center_left, center_top)
end

function PQS_WINDOW:load_player_settings(refresh_only)

	refresh_only = refresh_only or false
	
	-- ensure the user setting file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath..'/pqs/' .. self.settings_file, "r")
	if file == nil then 
		file = io.open(ShroudLuaPath..'/pqs/' .. self.settings_file, "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end
	
	self.player_settings = self.LIP().load(ShroudLuaPath..'/pqs/' .. self.settings_file)
	
	if (refresh_only == false) then 
		if (self.player_settings[player.name .. "|" .. self.title] == nil) then		
			self.player_settings[player.name .. "|" .. self.title] = {}
			self.player_settings[player.name .. "|" .. self.title]['left'] = self.left
			self.player_settings[player.name .. "|" .. self.title]['top'] = self.top
			self:save_player_settings()
		else
			self:moveTo(self.player_settings[player.name .. "|" .. self.title]['left'], self.player_settings[player.name .. "|" .. self.title]['top'])
		end
	end
end

function PQS_WINDOW:save_player_settings()
	self.LIP().save(ShroudLuaPath..'/pqs/' .. self.settings_file, self.player_settings)
end

function PQS_WINDOW:set_scale(new_scale)
	self.scale = new_scale
	self:rebuild()
	self:visible(false)
end

function PQS_WINDOW:set_zIndex(new_zIndex)
	self.zIndex = new_zIndex		
	self:rebuild()
end

function PQS_WINDOW:rebuild()
	self:rebuild_frame()
	
	if (self.rebuild_content ~= nil) then 
		self:rebuild_content()
	end
end

function PQS_WINDOW:close()
	self:visible(false)
end

function PQS_WINDOW:clear_elements()
	for element_key, element in pairs(self.elements) do
		if (element ~= nil) then 
			--print("clearing element: " .. element_key)
			element.element:remove()
		end
	end
	self.elements = {}
end

function PQS_WINDOW:rebuild_frame()

	self:clear_elements()

	self.width = self.original_width * self.scale
	self.height = self.original_height * self.scale

	local title_bar = ui.texture.add(0, 0, "pqs/modules/window-background.png", false, 2, self.original_width, 35)
	self:add_element("title-bar", title_bar, 0, 0)
	title_bar:zIndex(self.zIndex)
	
	local body = ui.texture.add(0, 0, "pqs/modules/window-background.png", false, 2, self.original_width, self.original_height)
	self:add_element("body", body, 0, 37)
	body:zIndex(self.zIndex)

	local title_label = ui.label.add(0, 0, self.original_width, 100, "<b>" .. self.title .. "</b>")
	self:add_element("title", title_label, 10, 7)
	title_label:zIndex(self.zIndex + self.controls_zIndex)
	
	local blank_tex = ui.texture.add(0, 0, "pqs/modules/tiny.png")

	if (self.hide_close_button == false) then 
		local close_btn = ui.guiButton.add(0, 0, blank_tex, "<b> X </b>", "close", 25, 25)
		self:add_element("close-button", close_btn, (self.original_width - 28), 4)
		close_btn:zIndex(self.zIndex + self.controls_zIndex)
		
		close_btn:onClick(function(x, y)
			if (self.window_system.mouse_over_window_key == self.window_system_key) then 
				self:close()
			end
		end)
	end
	
	if (self.hide_move_button == false) then 
		local move_btn = ui.guiButton.add(0, 0, blank_tex, "<b> M </b>", "move", 30, 25)
		move_btn:zIndex(self.zIndex + self.controls_zIndex)
		self:add_element("move-button", move_btn, (self.original_width - 60), 4)
		
		move_btn:onRepeat(function(x, y)
			if (self.window_system.mouse_over_window_key == self.window_system_key) then
				self.window_system:bring_to_front(self.window_system_key)
			end
		
			if (self.is_moving == false and self:is_top_window()) then 			
				self.is_moving = true
			end
		end)
	end
end

function PQS_WINDOW:add_element(element_name, element, relative_left, relative_top, relative_width, relative_height, font_size)

	font_size = font_size or 18
	relative_width = relative_width or element.rect.width
	relative_height = relative_height or element.rect.height
		
	local original_caption = nil

	if (element.caption ~= nil) then
		original_caption = element.caption
	end
	
	element:zIndex(self.zIndex + self.controls_zIndex)
	self.elements[element_name] = {
		element = element, 
		relative_left = relative_left, 
		relative_top = relative_top, 
		relative_width = relative_width,
		relative_height = relative_height, 
		original_caption = original_caption,
		font_size = font_size
	}
	element:visible(self.is_visible)

	-- adjust for scale
	element:moveTo(self.left + (relative_left * self.scale), self.top + (relative_top * self.scale))
	element:resizeTo(relative_width * self.scale, relative_height * self.scale)
	
	if (element.caption ~= nil) then
		element.caption = self:scale_text(original_caption, font_size, self.scale)
	end
	
	self.elements[element_name].window = self
end

function PQS_WINDOW:moveTo(left, top)
	self.left = left
	self.top = top
	
	for element_name, element in pairs(self.elements) do
		element.element:moveTo(self.left + (element.relative_left * self.scale), self.top + (element.relative_top  * self.scale))
	end
end

function PQS_WINDOW:visible(is_visible)
	self.is_visible = is_visible

	for element_name, element in pairs(self.elements) do	
		element.element:visible(self.is_visible)
	end

	if (is_visible) then 
		-- check to see if the window is off the right side
		if ((self.left + self.width > client.screen.width)) then
			self:load_player_settings(true)
			local shift_amount = (self.left + self.width) - client.screen.width
			self:moveTo(self.left - shift_amount, self.top)
			self:save_player_settings()
		end

		-- check to see if the window is off the bottom
		if ((self.top + (self.height + 40) > client.screen.height)) then
			self:load_player_settings(true)
			local shift_amount = (self.top + (self.height + 40)) - client.screen.height
			self:moveTo(self.left, self.top - (shift_amount))
			self:save_player_settings()
		end
	end
	
	if (self.on_visible ~= nil) then
		self:on_visible(self, self.is_visible)
	end
	
end

function PQS_WINDOW:scale_text(text_to_scale, font_size, scale)
	return "<size=" .. (font_size * scale) .. ">" .. text_to_scale .. "</size>"
end

return PQS_WINDOW