
local PQS_WINDOW_SYSTEM = {
	window_list = {},
	mouse_over_window_key = nil,
	last_ui_active = false,
	restore_vis = {}
}

function PQS_WINDOW_SYSTEM:initialize()
	ui.onUpdate(function()
		self:ui_active_check()
	end)

	ui.onMouseMove(function(button_data, mouse_x, mouse_y)
		-- button_data, key is button number, value is down/held/up	
		self.mouse_over_window_key = self:get_window_key_under_mouse(mouse_x, mouse_y)
		--print(self.mouse_over_window_key)
	end)
	
	ui.onMouseButton(function(click_type, button_index, mouse_x, mouse_y)		
		--"dblclicked/clicked/down/up"
		--print("mousebutton", click_type, button_index, mouse_x, mouse_y)
		--print(click_type, " | ", button_index)
		
		if (click_type == "clicked" and button_index == 1 and self.mouse_over_window_key ~= nil) then 
			self:bring_to_front(self.mouse_over_window_key)
		end		
	end)
end

function PQS_WINDOW_SYSTEM:reset_window_positions()
	for window_key, window in pairs(self.window_list) do
		window:move_to_center()
		window:load_player_settings(true)
		window:save_player_settings()
	end
end

function PQS_WINDOW_SYSTEM:get_window_key_under_mouse(mouse_x, mouse_y)
	local mouse_over_window_key = nil
	
	local highest_window = nil
	
	for window_key, window in pairs(self.window_list) do
		if window.is_visible and window:is_mouse_over(mouse_x, mouse_y) then
			if highest_window == nil or highest_window.zIndex < window.zIndex then 
				highest_window = window
			end
		end
	end

	if highest_window ~= nil then
		mouse_over_window_key = highest_window.window_system_key
	end
		
	return mouse_over_window_key
end

function PQS_WINDOW_SYSTEM:ui_active_check()
	local current_is_ui_active = ShroudIsUIActive()
				
	if self.last_ui_active == true and current_is_ui_active == false then -- hide
			
		for window_key, window in pairs(self.window_list) do
			if window ~= nil and window.is_visible then
				table.insert(self.restore_vis, window)
				window:visible(false)
			end
		end
		
	elseif self.last_ui_active == false and current_is_ui_active == true then -- restore
		for window_index, window in pairs (self.restore_vis) do
			window:visible(true)
		end
		self.restore_vis = {}
	end
	
	self.last_ui_active = current_is_ui_active	
end

function PQS_WINDOW_SYSTEM:window_exists(window_key)
	return (self.window_list[window_key] ~= nil)
end

function PQS_WINDOW_SYSTEM:get_window(window_key)
	local window = nil
	if self.window_list[window_key] ~= nil then 
		window = self.window_list[window_key]
	end
	return window
end

function PQS_WINDOW_SYSTEM:add_window(window_key, window)
	window.window_system = self
	window.window_system_key = window_key
	self.window_list[window_key] = window
	
	return window
end

function PQS_WINDOW_SYSTEM:remove_window(window_key)
	local window = nil
	if self.window_list[window_key] ~= nil then 
		window = self.window_list[window_key]
		self.window_list[window_key] = nil
	end
	return window
end

function PQS_WINDOW_SYSTEM:destroy_window(window_key)
	if self.window_list[window_key] ~= nil then 
		self.window_list[window_key]:clear_elements()
		self:remove_window(window_key)
	end
end

function PQS_WINDOW_SYSTEM:set_scale(new_scale)
	for window_key, window in pairs(self.window_list) do
		window:set_scale(new_scale)
	end
end

function PQS_WINDOW_SYSTEM:bring_to_front(bring_to_front_window_key)
	
	local behind_windows_zIndex = 1
	
	for window_key, window in pairs(self.window_list) do
		if window.is_visible == true then 
			if (window_key == bring_to_front_window_key) then			
				window:set_zIndex(100)
				self.mouse_over_window_key = window_key
			else
				window:set_zIndex(behind_windows_zIndex)
				behind_windows_zIndex = behind_windows_zIndex + 1
			end
		end
	end
end

function PQS_WINDOW_SYSTEM:close()
	for window_key, window in pairs(self.window_list) do
		window:visible(false)
	end
end

return PQS_WINDOW_SYSTEM